"""
scripts example to be launched from
"""
import os, copy
import numpy as np

import LaueTools
print("Using LaueToolsFolder: ", os.path.abspath(LaueTools.__file__))

import LaueTools.indexingSpotsSet as ISS
import LaueTools.matchingrate
import LaueTools.generaltools as GT
import LaueTools.findorient as FindO
import LaueTools.CrystalParameters as CP
import LaueTools.IOLaueTools as IOLT
import LaueTools.dict_LaueTools as DictLT
import LaueTools.indexingAnglesLUT as IAL
import LaueTools.lauecore as LAUE


def test_GetProximity():

    nexp = 3

    shufflenexp = np.random.permutation(nexp)

    TwicethetaChi = np.ascontiguousarray(
        np.random.randint(40, high=140, size=(2, 20)), dtype=np.float
    )

    data_theta = TwicethetaChi[0] / 2.0
    data_chi = TwicethetaChi[1] + 0.4

    exp_theta, exp_chi = data_theta[shufflenexp], data_chi[shufflenexp]

    res = matchingrate.getProximity(
        TwicethetaChi, exp_theta, exp_chi, angtol=0.5, proxtable=0, verbose=0, signchi=1
    )

    # allresidues, res, nb_in_res, len(allresidues), meanres, maxi

    print(res)
    return res, TwicethetaChi[0] / 2.0, exp_theta, TwicethetaChi[1], exp_chi


def test_GetProximity2():

    simuldata = np.array( [ [59.2164, -66.6041], [54.7398, -88.3875], [43.0763, -83.0983], [58.895, -88.7721], [39.3337, -65.6132], [52.4121, -58.2399], [47.6515, -100.4466], [47.9575, -83.2749], [44.0163, -64.8536], [56.4791, -57.2717], [48.434, -115.549], [52.2976, -101.5099], [28.9731, -75.0536], [53.191, -83.4718], [49.0293, -64.0107], [41.0373, -47.0408], [52.3338, -117.1192], [37.1386, -98.1813], [58.7918, -83.6928], [54.3839, -63.0704],
            [55.9659, -41.1327], [39.5682, -118.0816], [56.3583, -118.8114], [41.6482, -74.3025], [49.9809, -43.6082], [53.2592, -130.741], [44.4802, -120.2579], [49.9426, -100.9654], [38.9334, -47.8105], [54.7471, -41.6504], [56.4465, -132.6251], [49.6174, -122.6474], [21.8155, -95.1273], [57.3701, -73.2746], [46.408, -26.3132], [55.4884, -25.5038], [46.0081, -136.7264], [59.6475, -134.6084], [28.2351, -126.1327], [54.9433, -125.2734], [31.5495, -97.0398], [22.4182, -53.4955], [52.3414, -42.6513], [36.0373, -24.6759],
            [50.0743, -24.0159], [58.489, -23.5838], [35.0685, -129.9977], [43.26, -99.4789], [32.9485, -49.9361], [40.7146, -21.6059], [53.7612, -21.5865], [32.8305, -146.1445], [53.4871, -141.8828], [42.2965, -134.3526], [57.1994, -102.6829], [45.4058, -45.3993], [21.3158, -17.0592], [45.4556, -18.3182], [57.4384, -19.0229], [37.2146, -149.6308], [49.749, -139.236], [27.7533, -12.0696], [50.1926, -14.8127], [50.5507, -152.2943], [41.5523, -153.2806], [57.1871, -144.6661], [34.2669, -6.6991],
            [54.8492, -11.096], [53.1766, -154.6529], [45.7737, -157.0741], [12.2809, -166.6089], [40.6313, -1.0068], [59.3447, -7.1823], [52.0084, -4.6259], [55.7231, -157.0654], [49.8098, -160.985], [22.0916, -176.2917], [46.612, 4.9165], [43.6835, 1.9328], [54.5417, -2.2204], [42.4339, -165.4599], [58.1746, -159.525], [53.5979, -164.9811], [30.6904, 174.3774], [52.0051, 10.9533], [56.9817, 0.2302], [45.1098, -168.3128], [26.5826, 178.968], [57.086, -169.0252], [37.6256, 165.8261], [15.6917, 39.6224], [56.6694, 16.9722], [32.5476, 15.6023], [49.3926, 7.9288], [59.3144, 2.7178], [47.6427, -171.1734], [42.8967, 158.285], [22.5114, 46.4513], [35.8165, 19.2694],
            [50.0185, -174.0275], [34.3748, 169.985], [46.7627, 151.7992], [27.9587, 52.2656], [38.8316, 22.8625], [54.4341, 13.9733], [52.2268, -176.8611], [49.55, 146.2952], [55.2312, 100.1347], [32.2566, 57.1803], [41.5781, 26.3568], [54.2617, -179.6607], [40.4564, 161.9228], [51.5498, 141.6466], [35.6456, 61.3316], [44.0522, 29.7315], [58.7058, 19.9338], [56.121, 177.5861], [52.9876, 137.7171], [38.334, 64.8498], [46.2592, 32.9705], [57.8061, 174.8905], [44.9849, 154.9128], [54.0273, 134.3813], [40.4874, 67.8483], [48.2113, 36.0624],
            [59.3215, 172.2623], [54.7844, 131.5323], [42.2315, 70.4209], [49.9255, 39.0003], [41.3115, 160.6805], [48.2716, 148.9317], [55.3394, 129.0823], [43.6605, 72.6439], [51.4215, 41.7811], [55.7485, 126.9602], [44.8447, 74.5785], [52.7206, 44.4051], [44.3254, 156.0077], [50.6327, 143.8726], [56.0512, 125.1091], [45.8366, 76.2738], [53.8438, 46.875], [45.6099, 153.8468], [56.2755, 123.4835], [46.6758, 77.7691], [54.8115, 49.1957], [52.3274, 139.6002], [56.4416, 122.0467], [47.3926, 79.0961], [55.6426, 51.3733],
            [56.5638, 120.7692], [48.01, 80.2805], [56.3545, 53.4149], [53.549, 135.9821], [56.6529, 119.6269], [48.5461, 81.3433], [56.9627, 55.3281], [56.7166, 118.6003], [49.0149, 82.3015], [57.4811, 57.1208], [54.4355, 132.9019], [56.761, 117.6731], [49.4277, 83.1695], [57.9219, 58.8009], [56.7903, 116.8321], [49.7933, 83.9591], [58.2956, 60.376], [55.0832, 130.2622], [56.8081, 116.066], [50.1191, 84.6802], [58.6116, 61.8535], [56.8168, 115.3656], [50.4108, 85.3411], [58.8779, 63.2404],
            [55.5594, 127.9841], [56.8185, 114.7228], [59.1012, 64.5432], [56.8147, 114.1311], [56.8067, 113.5847], ]
    )

    expdata = np.ascontiguousarray(
        np.array(
            [
                [45.44929, -18.34126],
                [45.157076, -137.083181],
                [44.92114, 154.84827],
                [27.931473, 52.277335],
                [35.129301, -129.958303],
                [22.506704, 46.525391],
                [45.805636, -157.105433],
                [27.76149, -12.146467],
                [22.099332, -176.308319],
                [21.854819, -95.128858],
            ]
        ),
        dtype=np.int32,
    )

    TwicethetaChi = np.ascontiguousarray(simuldata, dtype=np.int32).T

    exp_theta, exp_chi = expdata[:, 0] / 2.0, expdata[:, 1]

    res = matchingrate.getProximity(
        TwicethetaChi, exp_theta, exp_chi, angtol=0.5, proxtable=0, verbose=0, signchi=1
    )

    # allresidues, res, nb_in_res, len(allresidues), meanres, maxi

    print(res)
    return res, TwicethetaChi[0] / 2.0, exp_theta / 2.0, TwicethetaChi[1], exp_chi


def Test_Build_and_Query_LUT():
    """
    test building and query of an angular look up table from a crystal
    given its lattice parameters

    """
    # test
    n = 3
    # compute LUT outside loop:
    hkl_all = GT.threeindices_up_to(n)

    if 1:  # filterharmonics:
        hkl_all = FindO.FilterHarmonics(hkl_all)

    #    Gstar_metric = CP.Gstar_from_directlatticeparams(2, 1, 4, 75, 90, 120)
    Gstar_metric = CP.Gstar_from_directlatticeparams(5.43, 5.43, 5.43, 90, 90, 90)
    # GenerateLookUpTable
    LUT = FindO.GenerateLookUpTable(hkl_all, Gstar_metric)

    print(FindO.QueryLUT(LUT, 14.01, 1))


def test_indexation_with_apriori_Matrices():
    """
    test of indexation + adding some selected matrices
    """
    folderdata = "./Examples/UO2/"
    filename_data = "dat_UO2_A163_2_0028_LT_0.cor"
    filename_data = os.path.join(folderdata, filename_data)
    StructureLabel = "UO2"

    nb_of_spots = 2000  # nb of spots to select for recognition
    nbspots_plot = 300  # nb of spots to select for plotting

    # READ .cor file
    mydata = IOLT.ReadASCIIfile(
        filename_data, col_2theta=0, col_chi=1, col_Int=-1, nblineskip=1
    )
    data_theta, data_chi, data_I = mydata

    print(data_I)
    print(data_theta)

    nbp = len(data_theta)
    # listofselectedpts=arange(nbofpeaks-1) # 1 for 1 header line
    # print data_I

    # --- Selection of peaks
    if nb_of_spots > 0:
        upto = min(nb_of_spots, nbp)
        sorted_int_index = np.argsort(data_I)[::-1][: upto + 1]
        print("Considering only %d most intense spots" % upto)

    elif nb_of_spots == -1:
        print("Considering all spots")
        sorted_int_index = np.argsort(data_I)[::-1]

    #    listofselectedpts = np.arange(len(sorted_int_index))
    #    indicespotmax=len(listofselectedpts)

    Theta = data_theta[sorted_int_index]
    Chi = data_chi[sorted_int_index]
    Intens = data_I[sorted_int_index]
    sorted_data = np.array([Theta, Chi, Intens]).T

    print(sorted_data[:10])

    # --- Rough interangular distance recognition

    # array of interangular distance of all points
    print("Calculating all angular distances\n")
    Tabledistance = GT.calculdist_from_thetachi(
        sorted_data[:, 0:2], sorted_data[:, 0:2]
    )

    ind_sorted_LUT_MAIN_CUBIC = [np.argsort(elem) for elem in FindO.LUT_MAIN_CUBIC]
    sorted_table_angle = []
    for k in list(range(len(ind_sorted_LUT_MAIN_CUBIC))):
        # print len(LUT_MAIN_CUBIC[k])
        # print len(ind_sorted_LUT_MAIN_CUBIC[k])
        sorted_table_angle.append(
            (FindO.LUT_MAIN_CUBIC[k])[ind_sorted_LUT_MAIN_CUBIC[k]]
        )

    # --- getOrientMatrices   with (spot_index_central,energy_max)
    spot_index = [ 0, 1, 21, 22, 24, ]  # one integer (starting from 0) or an array:array([1,8,9,5])
    energymax = 25
    nbmax_probed = (
        25
    )  # size of set of spots to calculate distance from central spot #spot_index
    nbspots_plot = 200  # for plot exp. data
    print("Using getOrientMatrices()")

    TrialsMatrix = (
        np.array(
            [
                [-0.99251179, 0.07104589, 0.09936213],
                [-0.01590782, -0.88248953, 0.47006294],
                [0.12126142, 0.46522732, 0.87684617],
            ]
        ),
        np.array(
            [
                [-0.49751201, -0.64141178, 0.58401432],
                [0.19466032, -0.7385631, -0.6454703],
                [0.84535393, -0.20737935, 0.49231143],
            ]
        ),
        np.eye(3),
    )

    TrialsMatrix = None

    # with nbbestplot very high  mama contain all matrices with matching rate above Nb_criterium

    n = 3  # for LUT
    latticeparams = DictLT.dict_Materials[StructureLabel][1]
    B = CP.calc_B_RR(latticeparams)

    mama, hhh = IAL.getOrientMatrices(
        spot_index,
        energymax,
        Tabledistance[:nbmax_probed, :nbmax_probed],
        Theta,
        Chi,
        n,
        B,
        LUT_tol_angle=0.5,
        MR_tol_angle=0.2,
        Minimum_Nb_Matches=40,
        key_material=StructureLabel,
        nbspots_plot=nbspots_plot,
        nbbestplot=20,
        plot=0,
        addMatrix=TrialsMatrix,
    )

    print("All %d matrices solutions" % len(mama))
    print(mama)
    print("Corresponding Scores")
    print(hhh)

    MAMA, SCO = ISS.MergeSortand_RemoveDuplicates(
        mama, hhh, 80, tol=0.0001, keep_only_equivalent=True
    )

    print(MAMA)
    print(SCO)


def test_indexation_generalRefunitCell_1():
    """
    test of indexation and unit cell refinement

    # lattice parameter : 2 1 4 75 90 120
    """
    B_input = np.array(
        [
            [0.58438544724018371, 0.62634243462603045, -0.040527337709274316],
            [0.0, 1.0352761804100832, -0.06698729810778066],
            [0.0, 0.0, 0.24999999999999997],
        ]
    )

    data_filecor = np.array(
        [
            [94.67479705, -1.77030003, 921.5841064, 877.0715942, 72.34869384],
            [73.69339752, 20.61630058, 572.9110107, 1218.955322, 72.01788330],
            [102.4393997, 8.985199928, 756.9523925, 755.0565795, 71.78106689],
            [80.51010131, 30.81780052, 381.7903137, 1118.804809, 69.76594543],
            [83.94760131, -11.9162998, 1077.943359, 1040.604248, 68.99935150],
            [63.29959869, 9.497099876, 754.0991210, 1386.598754, 68.80802154],  # 1 1 3
            [107.3254013, 19.63260078, 583.6207275, 661.9633789, 68.60917663],  # 0 1 3
            [84.57730102, 39.66809844, 180.0175933, 1058.541748, 64.89128112],
            [109.8495025, 29.49519920, 401.0426940, 589.9348754, 64.09487915],
            [111.0460968, -11.4846000, 1070.015625, 605.5197753, 62.07760238],
            [117.7855987, -1.76250004, 919.4577026, 488.4346923, 61.40348434],
            [70.78790283, -20.9498996, 1226.293212, 1267.149902, 61.09899902],
            [102.2435989, -20.2021999, 1213.381591, 745.6085205, 60.87683105],  # 1 1 2
            [122.3706970, 8.517900466, 761.3325805, 389.1836853, 59.37057495],
            [86.79679870, 47.08269882, -33.2585983, 1024.272338, 59.23816299],
            [110.6781005, 38.15859985, 208.5700988, 533.8281860, 59.13156890],
            [49.21139907, -1.77890002, 925.8856811, 1685.104736, 57.60239410],
            [91.72329711, -27.7292995, 1349.845825, 916.0936279, 57.46092987],  # 0 1 2
            [124.9682998, 18.72879981, 595.1906127, 304.0439147, 56.49592971],  # 2 1 3
        ]
    )

    mat_tofind = DictLT.dict_Rot["OrientSurf111"]

    a, b, c, alpha, beta, gamma = 2.0, 1.0, 4.0, 75.0, 90.0, 120.0
    #    latticeparams = a,b,c,alpha,beta,gamma
    B = CP.calc_B_RR([a, b, c, alpha, beta, gamma])
    print("B", B)
    # --- metric tensor
    #    Gstar_metric = CP.Gstar_from_directlatticeparams(a,b,c,alpha,beta,gamma)

    twiceTheta, Chi = data_filecor[:, :2].T

    thechi_exp = np.array([twiceTheta / 2.0, Chi]).T

    table_angdist = GT.calculdist_from_thetachi(thechi_exp, thechi_exp)

    ang_tol = 1
    n = 3
    spot_index = 5

    mat = IAL.matrices_from_onespot_new(
        spot_index, ang_tol, table_angdist, twiceTheta, Chi, n, B, verbose=1
    )

    index_matching = []
    for k, mat_sol in enumerate(mat[0]):
        if np.allclose(mat_sol, mat_tofind, atol=0.2):
            index_matching.append([mat[0][k], mat[1][k], mat[2][k]])

    print("These are the solutions !...")
    print(index_matching)


def test_indexation_generalRefunitCell_2():
    """
    test of indexation and unit cell refinement

    [2,1,4,90,90,90] lattice parameter oriented with orientsurface111
    """
    data_filecor = np.array(
        [
            [79.98930358, -0.01720000, 896.3380737, 1100.323608, 59.85998535],
            [72.14089965, -7.87809991, 1015.881286, 1227.356567, 59.33337783],  # 2 1 4
            [85.72489929, 7.331699848, 784.4506835, 1013.795471, 58.13841629],  # 2 1 3
            [61.93239974, -15.9895000, 1143.677978, 1422.501708, 55.33495330],  # 2 1 5
            [61.84360122, 15.91269969, 653.3759155, 1426.837646, 55.26354980],  # 2 1 2
            [89.83450317, -14.0176000, 1111.113281, 949.4226074, 55.25106430],  # 1 1 4
            [89.75630187, 14.00380039, 679.5944213, 953.0634155, 55.21326446],  # 3 1 4
            [95.69380187, -6.89139986, 999.4022216, 860.4885864, 54.51536941],  # 2 1 6
            [82.30560302, -21.1537990, 1229.820800, 1071.203369, 54.36821365],  # 3 1 5
            [68.30660247, 22.44009971, 542.4238891, 1319.939941, 54.00218200],  # 3 1 3
            [52.94110107, 8.390600204, 772.7620849, 1601.591430, 53.40956115],
        ]
    )

    mat_tofind = DictLT.dict_Rot["OrientSurf111"]

    a, b, c, alpha, beta, gamma = 2.0, 1.0, 4.0, 90.0, 90.0, 90.0
    #    latticeparams = a,b,c,alpha,beta,gamma
    B = CP.calc_B_RR([a, b, c, alpha, beta, gamma])
    # --- metric tensor
    #    Gstar_metric = CP.Gstar_from_directlatticeparams(a,b,c,alpha,beta,gamma)

    twiceTheta, Chi = data_filecor[:, :2].T

    thechi_exp = np.array([twiceTheta / 2.0, Chi]).T

    table_angdist = GT.calculdist_from_thetachi(thechi_exp, thechi_exp)

    ang_tol = 1
    n = 4
    spot_index = 1

    mat = IAL.matrices_from_onespot_new(
        spot_index, ang_tol, table_angdist, twiceTheta, Chi, n, B, verbose=0
    )

    print("mat", mat)

    index_matching = []
    for k, mat_sol in enumerate(mat[0]):
        if np.allclose(mat_sol, mat_tofind, atol=0.2):
            index_matching.append([mat[0][k], mat[1][k], mat[2][k]])

    print("These are the solutions !...")
    print(index_matching)


def Test_getOrients_AnglesMatching():
    """
    test for the classical method implemented in Common_index_Method()
    """
    readfile = None

    spot_index_central = [0, 1, 2]
    key_material = "UO2"
    emax = 22

    nbmax_probed = 8000000

    angleTolerance_LUT = 0.1
    MatchingRate_Angle_Tol = 0.001
    Matching_Threshold_Stop = (50.0)  # minimum to stop the loop for searching potential solution

    # create a fake data file of randomly oriented crystal
    if readfile is None:
        outputfilename = "toto"
        nbgrains = 3
        removespots = [0, 0, 0]  # , 0, 3, 10, 2, 1, 5]
        addspots = 0
        #        removespots = None
        #        addspots = None
        file_to_index = createFakeData(key_material,
                                        nbgrains,
                                        outputfilename=outputfilename,
                                        removespots=removespots,
                                        addspots=addspots)
    else:
        file_to_index = readfile

    # read data
    data_theta, Chi, posx, posy, dataintensity, detectorparameters = IOLT.readfile_cor(
        file_to_index)[1:]
    tth = 2 * data_theta

    exp_data = np.array([tth, Chi])

    thechi_exp = np.array([data_theta, Chi]).T

    table_angdist = GT.calculdist_from_thetachi(thechi_exp, thechi_exp)

    latticeparams = DictLT.dict_Materials[key_material][1]
    B = CP.calc_B_RR(latticeparams)

    nbmax_probed = min(nbmax_probed, len(data_theta))

    matrix, score, Threshold_reached = IAL.getOrients_AnglesLUT(spot_index_central,
                                                    table_angdist[:nbmax_probed, :nbmax_probed],
                                                    tth,
                                                    Chi,
                                                    Matching_Threshold_Stop=Matching_Threshold_Stop,
                                                    n=3,  # up  to (332)
                                                    B=B,  # for cubic
                                                    LUT=None,
                                                    angleTolerance_LUT=angleTolerance_LUT,
                                                    MatchingRate_Angle_Tol=MatchingRate_Angle_Tol,
                                                    key_material=key_material,
                                                    emax=emax,
                                                    verbose=0)
    #
    #    bestmat, stats_res = ISS.MergeSortand_RemoveDuplicates(bestmat, stats_res, Nb_criterium,
    #                                                           tol=0.0001, allpermu=True)

    print("B", B)
    print("matrix", matrix)
    return matrix, score, detectorparameters  # , exp_data


def plotgrains(listmat, key_material, emax, file_to_index="toto.cor"):
    """
    plot grains spots and data from orientmatrix list up to 9
    """
    # read data
    data_theta, Chi, posx, posy, dataintensity, detectorparameters = IOLT.readfile_cor(
        file_to_index
    )[1:]
    tth_exp = 2 * data_theta
    Chi_exp = Chi

    EXTINCTION = DictLT.dict_Materials[key_material][2]

    nb_matrices = 0
    all_tthchi = []
    for mat in listmat:

        print("mat in plotgrains", mat)

        grain = [np.eye(3), EXTINCTION, mat, key_material]

        (Twicetheta_theo,
            Chi_theo,
            Miller_ind_theo,
            posx_theo,
            posy_theo,
            Energy_theo,
        ) = LAUE.SimulateLaue(grain, 5, emax, detectorparameters)

        all_tthchi.append([Twicetheta_theo, Chi_theo])

        nb_matrices += 1
        if nb_matrices == 9:
            break

    print("nb_of_matrices to plot", nb_matrices)
    if nb_matrices == 1:
        codefigure = 111
    if nb_matrices == 2:
        codefigure = 211
    if nb_matrices in (3, 4):
        codefigure = 221
    if nb_matrices in (5, 6):
        codefigure = 321
    if nb_matrices in (7, 8, 9):
        codefigure = 331

    index_fig = 0

    #    ax.set_xlim((35, 145))
    #    ax.set_ylim((-45, 45))

    dicocolor = {0: "k", 1: "r", 2: "g", 3: "b", 4: "c", 5: "m"}
    nbcolors = len(dicocolor)

    import pylab as p

    # theo spots
    for i_mat in list(range(nb_matrices)):
        p.subplot(codefigure)

        if file_to_index is not None:
            # all exp spots
            p.scatter(tth_exp, Chi_exp, s=40, c="w", marker="o", faceted=True, alpha=0.5)

        # simul spots
        p.scatter(
            all_tthchi[i_mat][0],
            all_tthchi[i_mat][1],
            c=dicocolor[(i_mat + 1) % nbcolors],
            faceted=False,
        )

        if index_fig < nb_matrices:
            index_fig += 1
            codefigure += 1
            if nb_matrices == 9:
                break

    p.show()


# --- ------------ Simulate fake data
def IofE(Energy, Amplitude=50000, powerp=-1, offset=-5):
    """
    returns simulated intensity of bragg spot as a function of Energy only
    """
    return 1.0 * Amplitude * np.power(Energy + offset, powerp * 1.0)


def Grain_to_string(grains):
    """
    convert each 4 elements grain parameters contained in grains list to list of string
    """
    liststr = ["mat, EXTINCTION, orientmat, ELEMENT"]
    for k, grain in enumerate(grains):
        mat, EXTINCTION, orientmat, ELEMENT = grain

        matstr = ISS.matrix_to_string(mat)
        orientmatstr = ISS.matrix_to_string(orientmat)

        liststr.append("Grain G % d" % k)
        for elem in matstr:
            liststr.append(elem)
        liststr.append(EXTINCTION)
        for elem in orientmatstr:
            liststr.append(elem)
        liststr.append(ELEMENT)

    return liststr


def randomdetectorparameters():
    """
    generate random detector parameters
    """
    randomdet = (np.random.rand() - 0.5) * 2 * 3 + 70.0
    randomxcen = (np.random.rand() - 0.5) * 2 * 300 + 1024
    randomycen = (np.random.rand() - 0.5) * 2 * 300 + 1024
    randomxbet = (np.random.rand() - 0.5) * 2 * 2 + 0
    randomxgam = (np.random.rand() - 0.5) * 2 * 4 + 0

    detectorparameters = [randomdet, randomxcen, randomycen, randomxbet, randomxgam]

    return detectorparameters


def createFakeData(
    key_material,
    nbgrains,
    emin=5,
    emax=25,
    outputfilename=None,
    removespots=None,
    addspots=None,
    twins=0,
):
    """
    create fake data with nbgrains grains and random detecgtor parameters

    removespots    : nb of spots to remove (randomly) in each grains
    addspots        : number of random experimental spots to add
    """
    if not outputfilename:
        outputfilename = "testSirandom"
        print("%s.cor has be written" % outputfilename)
    #    detectorparameters = [70, 1024, 1024, 0, 0]
    detectorparameters = randomdetectorparameters()

    # generate .cor file
    simul_matrices = []
    grains = []
    Twicetheta, Chi, posx, posy, dataintensity = [], [], [], [], []

    EXTINCTION = DictLT.dict_Materials[key_material][2]

    if removespots is not None:
        if len(removespots) != nbgrains:
            raise ValueError("removespots needs as many integers than nbgrains!")

    for k in list(range(nbgrains)):
        orientmat = GT.randomRotationMatrix()
        simul_matrices.append(orientmat)

    if twins >= 1:
        if twins <= nbgrains / 2:
            copysimul_matrices = copy.copy(simul_matrices)
            for twin in list(range(0, twins + 1, 2)):
                print("twin", twin)
                simul_matrices[twin + 1] = np.dot(
                    np.dot(copysimul_matrices[twin], DictLT.dict_Vect["sigma3_1"]),
                    GT.matRot([1, 2, 3], 0),
                )
        else:
            raise ValueError("nb of twins must not be higher than half of nbgrains")

    for k, orientmat in enumerate(simul_matrices):

        grains.append([np.eye(3), EXTINCTION, orientmat, key_material])

        print("orientmat", orientmat)

        # Twicetheta, Chi, Miller_ind, posx, posy, Energy
        (Twicetheta1, Chi1, Miller_ind1, posx1, posy1, Energy1) = LAUE.SimulateLaue(
            grains[k], emin, emax, detectorparameters, removeharmonics=1
        )

        if removespots is not None:
            # TODO : to simplify
            nbspots_init = len(Twicetheta1)
            print("nbspots_init", nbspots_init)
            toremoveind = np.random.randint(nbspots_init, size=removespots[k])

            # remove duplicate
            toremoveind = list(set(toremoveind))
            print("nb of removed spots", len(toremoveind))
            Twicetheta1 = np.delete(Twicetheta1, toremoveind, axis=0)
            Chi1 = np.delete(Chi1, toremoveind, axis=0)
            posx1 = np.delete(posx1, toremoveind, axis=0)
            posy1 = np.delete(posy1, toremoveind, axis=0)
            Energy1 = np.delete(Energy1, toremoveind, axis=0)

            Miller_ind1 = np.delete(Miller_ind1, toremoveind, axis=0)

        #            dataintensity1 = (20000 + k) * np.ones(len(Twicetheta1))
        dataintensity1 = IofE(Energy1)

        Twicetheta = np.concatenate((Twicetheta, Twicetheta1))
        Chi = np.concatenate((Chi, Chi1))
        posx = np.concatenate((posx, posx1))
        posy = np.concatenate((posy, posy1))
        dataintensity = np.concatenate((dataintensity, dataintensity1))

        k += 1

    if twins:
        # need to delete duplicates
        Twicetheta, Chi, tokeep = GT.removeClosePoints_2(
            Twicetheta, Chi, dist_tolerance=0.01
        )
        posx = np.take(posx, tokeep)
        posy = np.take(posy, tokeep)
        dataintensity = np.take(dataintensity, tokeep)

    if addspots is not None:

        posx1 = (2048) * np.random.random_sample(addspots)
        posy1 = (2048) * np.random.random_sample(addspots)
        dataintensity1 = (65000) * np.random.random_sample(addspots)

        from LaueGeometry import calc_uflab

        Twicetheta1, Chi1 = calc_uflab(posx1, posy1, detectorparameters)

        Twicetheta = np.concatenate((Twicetheta, Twicetheta1))
        Chi = np.concatenate((Chi, Chi1))
        posx = np.concatenate((posx, posx1))
        posy = np.concatenate((posy, posy1))
        dataintensity = np.concatenate((dataintensity, dataintensity1))

    # write fake .cor file
    strgrains = ["simulated pattern with two grains"]
    strgrains += Grain_to_string(grains)

    # sorting data by decreasing intensity:
    intensitysortedind = np.argsort(dataintensity)[::-1]
    arraydata = np.array([Twicetheta, Chi, posx, posy, dataintensity]).T

    sortedarraydata = arraydata[intensitysortedind]

    Twicetheta, Chi, posx, posy, dataintensity = sortedarraydata.T

    IOLT.writefile_cor(
        outputfilename,
        Twicetheta,
        Chi,
        posx,
        posy,
        dataintensity,
        param=detectorparameters,
        comments=strgrains,
    )

    fullname = outputfilename + ".cor"
    print("Fake data have been saved in %s" % fullname)
    return fullname


# --- -----------------  Tests
def example_angularresidues_np(ang_tol):

    detectorparameters = {}
    detectorparameters["kf_direction"] = "Z>0"
    detectorparameters["detectorparameters"] = [70, 1024, 1024, 0, -1]
    detectorparameters["detectordiameter"] = 170
    detectorparameters["pixelsize"] = 0.079
    detectorparameters["dim"] = (2048, 2048)

    test_Matrix = np.eye(3)
    # random data
    expdata = np.array(
        [
            [59.2164, -66.6041],
            [54.7398, -88.3875],
            [43.0763, -83.0983],
            [58.895, -88.7721],
            [39.3337, -65.6132],
            [52.4121, -58.2399],
            [47.6515, -100.4466],
            [47.9575, -83.2749],
            [44.0163, -64.8536],
            [56.4791, -57.2717],
            [48.434, -115.549],
            [52.2976, -101.5099],
            [28.9731, -75.0536],
            [53.191, -83.4718],
            [49.0293, -64.0107],
            [41.0373, -47.0408],
            [52.3338, -117.1192],
            [37.1386, -98.1813],
            [58.7918, -83.6928],
            [54.3839, -63.0704],
            [55.9659, -41.1327],
            [39.5682, -118.0816],
            [56.3583, -118.8114],
            [41.6482, -74.3025],
            [49.9809, -43.6082],
            [53.2592, -130.741],
            [44.4802, -120.2579],
            [49.9426, -100.9654],
            [38.9334, -47.8105],
        ]
    )

    twicetheta_data, chi_data = expdata.T

    AngRes = matchingrate.Angular_residues_np(
        test_Matrix,
        twicetheta_data,
        chi_data,
        ang_tol=ang_tol,
        key_material="CdTe",
        emin=5,
        emax=40,
        ResolutionAngstrom=False,
        detectorparameters=detectorparameters,
    )

    if AngRes is not None:
        allresidues, res, nb_in_res, nb_of_theospots, meanres, maxi = AngRes

        matchringrate = 1.0 * nb_in_res / nb_of_theospots * 100.0
        print("nb of theo spots", nb_of_theospots)
        print("matching rate is %.2f" % matchringrate)


def test_IAM_real():
    """
    test refinement and multigrains indexing with class wit angles LUT indexing technique
    """
    #    readfile = 'test6.cor'
    #    readfile = 'test7.cor'
    #    readfile = 'test10.cor'
    import time, os

    readfile = os.path.join("./Examples/UO2", "dat_UO2_A163_2_0028_LT_0.cor")

    key_material = "UO2"
    emin = 5
    emax = 22

    nbGrainstoFind = 3  # 'max'#'1 #'max'

    MatchingRate_Threshold_Si = 30  # percent for simulated Si
    dictimm_Si = {
        "sigmagaussian": (0.5, 0.5, 0.5),
        "Hough_init_sigmas": (1, 0.7),
        "Hough_init_Threshold": 1,
        "rank_n": 20,
        "useintensities": 0,
    }

    MatchingRate_Threshold_W = 15  # percent W
    dictimm_W = {
        "sigmagaussian": (0.5, 0.5, 0.5),
        "Hough_init_sigmas": (1, 0.7),
        "Hough_init_Threshold": 1,
        "rank_n": 40,
        "useintensities": 0,
    }

    MatchingRate_Threshold = MatchingRate_Threshold_W
    dictimm = dictimm_W

    database = None
    file_to_index = readfile

    #    t0 = time.time()
    #
    #    DataSet = spotsset()
    #
    #    DataSet.IndexSpotsSet(file_to_index, key_material, emin, emax, dictimm, database,
    #                          IMM=True,
    #                          nbGrainstoFind=nbGrainstoFind)
    #
    #    tf1 = time.time()
    #    print "imageMatching execution time %.3f sec." % (tf1 - t0)
    #
    #    DataSet.plotallgrains()

    # IAM technique
    print("\n\n\n\n\n")
    print("ANGLES LUT ---------------------------------------------")
    print("\n\n\n\n\n")

    dict_params = {
        "MATCHINGRATE_THRESHOLD_IAL": 60,
        "MATCHINGRATE_ANGLE_TOL": 0.2,
        "NBMAXPROBED": 10,
        "central spots indices": 10,
    }

    t0_2 = time.time()
    DataSet_2 = spotsset()

    DataSet_2.IndexSpotsSet(
        file_to_index,
        key_material,
        emin,
        emax,
        dict_params,
        database,
        IMM=False,
        MatchingRate_List=[20, 20, 20],
        nbGrainstoFind=nbGrainstoFind,
    )

    tf2 = time.time()
    print("Angles LUT execution time %.3f sec." % (tf2 - t0_2))

    DataSet_2.plotallgrains()
    return DataSet_2  # , DataSet_2


def testIAM():
    """
    test indexing by angles Matching
    """
    readfile = None

    spot_index_central = [0, 1, 2]
    key_material = "Cu"
    emin = 5
    emax = 22

    nbmax_probed = 8000000
    nbGrainstoFind = 10

    angleTolerance_LUT = 0.1
    MatchingRate_Angle_Tol = 0.001
    Matching_Threshold_Stop = (
        50.0
    )  # minimum to stop the loop for searching potential solution

    # create a fake data file of randomly oriented crystal
    if readfile is None:
        outputfilename = "toto"
        nbgrains = nbGrainstoFind
        removespots = [0] * nbgrains
        addspots = 3
        #        removespots = None
        #        addspots = None
        file_to_index = createFakeData(
            key_material,
            nbgrains,
            emin=emin,
            emax=emax,
            outputfilename=outputfilename,
            removespots=removespots,
            addspots=addspots,
        )
    else:
        file_to_index = readfile

    # read data
    data_theta, Chi, posx, posy, dataintensity, detectorparameters = IOLT.readfile_cor(
        file_to_index
    )[1:]
    tth = 2 * data_theta

    exp_data = np.array([tth, Chi])

    thechi_exp = np.array([data_theta, Chi]).T

    table_angdist = GT.calculdist_from_thetachi(thechi_exp, thechi_exp)

    latticeparams = DictLT.dict_Materials[key_material][1]
    B = CP.calc_B_RR(latticeparams)

    nbmax_probed = min(nbmax_probed, len(data_theta))
    # ---------------------------------------------------------------
    import time

    t0_2 = time.time()
    DataSet_2 = spotsset()

    database = None
    dictimm = None
    DataSet_2.IndexSpotsSet(
        file_to_index,
        key_material,
        emin,
        emax,
        dictimm,
        database,
        IMM=False,
        MatchingRate_List=[20, 20, 20],
        nbGrainstoFind=nbGrainstoFind,
        verbose=0,
    )

    tf2 = time.time()
    print("Angles LUT execution time %.3f sec." % (tf2 - t0_2))

    DataSet_2.plotallgrains()


# --- CuVia analysis
def testtwin(
    fileindex=2301,
    nbGrainstoFind=3,
    dirname="/home/micha/LaueProjects/CuVia/Carto",
    dirtowrite=None,
):
    """
    test twin
    """
    if dirtowrite is None:
        dirtowrite = dirname

    #     spot_index_central = [0, 1, 2]
    key_material = "Si"
    emin = 5
    emax = 25

    file_to_index = os.path.join(dirname, "Si_TSV.cor")

    t0_2 = time.time()
    DataSet_Si = spotsset()

    database = None

    dict_params = {
        "MATCHINGRATE_THRESHOLD_IAL": 60,
        "MATCHINGRATE_ANGLE_TOL": 0.2,
        "NBMAXPROBED": 10,
        "central spots indices": 10,
        "AngleTolLUT": 0.5,
        "UseIntensityWeights": True,
        "MinimumNumberMatches": 10,
    }

    DataSet_Si.IndexSpotsSet(
        file_to_index,
        key_material,
        emin,
        emax,
        dict_params,
        database,
        IMM=False,
        MatchingRate_List=[20, 20, 20],
        angletol_list=[0.5, 0.2],
        nbGrainstoFind=1,
        verbose=0,
    )

    tf2 = time.time()
    print("Angles LUT execution time %.3f sec." % (tf2 - t0_2))

    DataSet_Si.plotallgrains()

    # ind, 2theta, chi, posx, posy, int
    dataSubstrate = DataSet_Si.getSpotsFamilyExpData(0).T

    DataSet_Cu = spotsset()
    DataSet_Cu.importdatafromfile(file_to_index)

    DataSet_Cu.purgedata(dataSubstrate[1:3], dist_tolerance=0.5)

    key_material = "Cu"
    dict_params = {
        "MATCHINGRATE_THRESHOLD_IAL": 40,
        "MATCHINGRATE_ANGLE_TOL": 0.2,
        "NBMAXPROBED": 20,
        "central spots indices": 3,
        "AngleTolLUT": 0.5,
        "UseIntensityWeights": True,
        "MinimumNumberMatches": 10,
    }

    dictMat = {}
    dictMR = {}
    dictNB = {}
    dictstrain = {}
    dictRes = (dictMat, dictMR, dictNB, dictstrain)

    nstart = fileindex
    nend = fileindex

    for k in list(range(nstart, nend + 1)):  # 92 - 1708
        file_to_index = os.path.join(dirname, "TSVCU_%04d.cor" % k)

        print("\n\nINDEXING    file : %s\n\n" % file_to_index)

        DataSet_Cu = spotsset()
        DataSet_Cu.importdatafromfile(file_to_index)

        DataSet_Cu.purgedata(dataSubstrate[1:3], dist_tolerance=0.5)

        # init res dict
        DataSet_Cu.dict_grain_matrix = [0, 0, 0]
        DataSet_Cu.dict_grain_matching_rate = [[-1, -1], [-1, -1], [-1, -1]]

        # filling results
        dictMat[k] = [0, 0, 0]
        dictMR[k] = [-1, -1, -1]
        dictNB[k] = [-1, -1, -1]
        dictstrain[k] = [0, 0, 0]
        previousResults = None

        if k > nstart and dictMat[k - 1][0] is not 0:
            addMatrix = dictMat[k - 1][0]
            previousResults = addMatrix, dictMR[k - 1][0], dictNB[k - 1][0]

        DataSet_Cu.IndexSpotsSet(
            file_to_index,
            key_material,
            emin,
            emax,
            dict_params,
            database,
            checkSigma3=True,
            use_file=0,
            IMM=False,
            angletol_list=[0.5, 0.5],
            MatchingRate_List=[10, 10, 10],
            nbGrainstoFind=nbGrainstoFind,
            verbose=0,
            previousResults=previousResults,
        )

        for nbgrain in list(range(nbGrainstoFind)):
            dictMat[k][nbgrain] = DataSet_Cu.dict_grain_matrix[nbgrain]
            dictMR[k][nbgrain] = DataSet_Cu.dict_grain_matching_rate[nbgrain][1]
            dictNB[k][nbgrain] = DataSet_Cu.dict_grain_matching_rate[nbgrain][0]
            dictstrain[k][nbgrain] = DataSet_Cu.dict_grain_devstrain[nbgrain]

        # intermediate saving
        if (k % 100) == 0:
            #            dictRes = dictMat, dictMR, dictNB
            filepickle = open("dictCuViaaddMatrix%04d_%04d" % (nstart, nend), "w")
            pickle.dump(dictRes, filepickle)
            filepickle.close()

    filepickle = open("dictCuViaaddMatrix%04d_%04d" % (nstart, nend), "w")
    pickle.dump(dictRes, filepickle)
    filepickle.close()

    #        DataSet_Cu.plotallgrains()
    return dictRes


#    return DataSet_Si, DataSet_Cu


def SiCu(nstart, nend):
    """
    shortcut of file serie indexation of Si and Cu grains
    """
    return test_SiCu(nstart=nstart, nend=nend)


def test_SiCu(nstart=92, nend=1707):
    """
    test to index a file series with Si pattern and Cu pattern
    with starting and ending image index as arguments
    """
    import pickle

    spot_index_central = [0, 1, 2]
    key_material = "Si"
    emin = 5
    emax = 25
    nbGrainstoFind = 1

    import time, os

    file_to_index = os.path.join(
        "/home/micha/LaueProjects/CuVia/filecor", "dat_TSVCU_0092.cor"
    )

    t0_2 = time.time()
    DataSet_Si = spotsset()

    database = None

    dict_params = {
        "MATCHINGRATE_THRESHOLD_IAL": 60,
        "MATCHINGRATE_ANGLE_TOL": 0.2,
        "NBMAXPROBED": 10,
        "central spots indices": 10,
    }

    DataSet_Si.IndexSpotsSet(
        file_to_index,
        key_material,
        emin,
        emax,
        dict_params,
        database,
        IMM=False,
        MatchingRate_List=[20, 20, 20],
        angletol_list=[0.5, 0.2],
        nbGrainstoFind=nbGrainstoFind,
        verbose=0,
    )

    tf2 = time.time()
    print("Angles LUT execution time %.3f sec." % (tf2 - t0_2))

    DataSet_Si.plotallgrains()

    # ind, 2theta, chi, posx, posy, int
    dataSubstrate = DataSet_Si.getSpotsFamilyExpData(0).T

    DataSet_Cu = spotsset()
    DataSet_Cu.importdatafromfile(file_to_index)

    DataSet_Cu.purgedata(dataSubstrate[1:3], dist_tolerance=0.5)

    key_material = "Cu"
    nbGrainstoFind = 3
    dict_params = {
        "MATCHINGRATE_THRESHOLD_IAL": 40,
        "MATCHINGRATE_ANGLE_TOL": 0.2,
        "NBMAXPROBED": 20,
        "central spots indices": 3,
    }

    dictMat = {}
    dictMR = {}
    dictNB = {}

    for k in list(range(nstart, nend + 1)):  # 92 - 1708
        file_to_index = os.path.join(
            "/home/micha/LaueProjects/CuVia/filecor", "TSVCU_%04d.cor" % k
        )

        print("\n\nINDEXING    file : %s\n\n" % file_to_index)

        DataSet_Cu = spotsset()
        DataSet_Cu.importdatafromfile(file_to_index)

        DataSet_Cu.purgedata(dataSubstrate[1:3], dist_tolerance=0.5)

        # init res dict
        DataSet_Cu.dict_grain_matrix = [0, 0, 0]
        DataSet_Cu.dict_grain_matching_rate = [[-1, -1], [-1, -1], [-1, -1]]

        DataSet_Cu.IndexSpotsSet(
            file_to_index,
            key_material,
            emin,
            emax,
            dict_params,
            database,
            use_file=0,
            IMM=False,
            angletol_list=[0.5, 0.5],
            MatchingRate_List=[10, 10, 10],
            nbGrainstoFind=nbGrainstoFind,
            verbose=0,
        )
        # filling results
        dictMat[k] = [0, 0, 0]
        dictMR[k] = [-1, -1, -1]
        dictNB[k] = [-1, -1, -1]
        for nbgrain in list(range(nbGrainstoFind)):
            dictMat[k][nbgrain] = DataSet_Cu.dict_grain_matrix[nbgrain]
            dictMR[k][nbgrain] = DataSet_Cu.dict_grain_matching_rate[nbgrain][1]
            dictNB[k][nbgrain] = DataSet_Cu.dict_grain_matching_rate[nbgrain][0]

        # intermediate saving
        dictRes = dictMat, dictMR, dictNB
        if (k % 100) == 0:
            filepickle = open("dictCuVia%04d_%04d" % (nstart, nend), "w")
            pickle.dump(dictRes, filepickle)
            filepickle.close()

    filepickle = open("dictCuVia", "w")
    pickle.dump(dictRes, filepickle)
    filepickle.close()

    #        DataSet_Cu.plotallgrains()
    return dictRes


#    return DataSet_Si, DataSet_Cu


def index_fileseries(
    fileindexrange,
    nbGrainstoFind=1,
    dirname="/home/micha/LaueProjects/CuVia/Carto",
    dirtowrite=None,
    saveObject=0,
):
    """
    This is still an example specific to CuVia (Cu and Si)
    for CuVia Jul 11
    nstart = 1708
    nend = 3323
    """
    nstart, nend = fileindexrange

    if dirtowrite is None:
        dirtowrite = dirname

    spot_index_central = [0, 1, 2]
    key_material = "Si"
    emin = 5
    emax = 25
    nbGrainstoFind_Si = 1

    import time

    file_to_index = os.path.join(dirname, "Si_TSV.cor")

    t0_2 = time.time()
    DataSet_Si = spotsset()

    database = None

    dict_params = {
        "MATCHINGRATE_THRESHOLD_IAL": 60,
        "MATCHINGRATE_ANGLE_TOL": 0.2,
        "NBMAXPROBED": 10,
        "central spots indices": 10,
        "AngleTolLUT": 0.5,
        "UseIntensityWeights": True,
        "MinimumNumberMatches": 10,
    }

    DataSet_Si.IndexSpotsSet(
        file_to_index,
        key_material,
        emin,
        emax,
        dict_params,
        database,
        IMM=False,
        MatchingRate_List=[20, 20, 20],
        angletol_list=[0.5, 0.2],
        nbGrainstoFind=nbGrainstoFind_Si,
        verbose=0,
    )

    tf2 = time.time()
    print("Angles LUT execution time %.3f sec." % (tf2 - t0_2))

    # DataSet_Si.plotallgrains() # to comment when mutliprocessing

    # ind, 2theta, chi, posx, posy, int
    dataSubstrate = DataSet_Si.getSpotsFamilyExpData(0).T

    key_material = "Cu"
    nbGrainstoFind_Cu = nbGrainstoFind
    dict_params = {
        "MATCHINGRATE_THRESHOLD_IAL": 40,
        "MATCHINGRATE_ANGLE_TOL": 0.2,
        "NBMAXPROBED": 20,
        "central spots indices": 3,
        "AngleTolLUT": 0.5,
        "UseIntensityWeights": True,
        "MinimumNumberMatches": 10,
    }

    dictMat = {}
    dictMR = {}
    dictNB = {}
    dictstrain = {}
    dictspots = {}

    if saveObject:
        dict_spotssetObj = {}

    #    dictRes = {'dictMat':dictMat,
    #               'dictMR':dictMR,
    #               'dictNB':dictNB,
    #               'dictstrain':dictstrain}
    dictRes = dictMat, dictMR, dictNB, dictstrain, dictspots

    if saveObject:
        todump = dictRes, dict_spotssetObj
    else:
        todump = dictRes

    for k in list(range(nstart, nend + 1)):  # 1708-3323
        file_to_index = os.path.join(dirname, "TSVCU_%04d.cor" % k)

        print("\n\nINDEXING    file : %s\n\n" % file_to_index)

        DataSet_Cu = spotsset()
        DataSet_Cu.importdatafromfile(file_to_index)

        DataSet_Cu.purgedata(dataSubstrate[1:3], dist_tolerance=0.5)

        # TODO usefull ?
        # init res dict
        DataSet_Cu.dict_grain_matrix = [0 for kk in list(range(nbGrainstoFind_Cu))]
        DataSet_Cu.dict_grain_matching_rate = [
            [-1, -1] for kk in list(range(nbGrainstoFind_Cu))
        ]
        DataSet_Cu.dict_grain_devstrain = [0 for kk in list(range(nbGrainstoFind_Cu))]

        # preparing dicts of results
        dictMat[k] = [0 for kk in list(range(nbGrainstoFind_Cu))]
        dictMR[k] = [-1 for kk in list(range(nbGrainstoFind_Cu))]
        dictNB[k] = [-1 for kk in list(range(nbGrainstoFind_Cu))]
        dictstrain[k] = [0 for kk in list(range(nbGrainstoFind_Cu))]
        previousResults = None

        if k > nstart and dictMat[k - 1][0] is not 0:
            addMatrix = dictMat[k - 1][0]
            previousResults = addMatrix, dictMR[k - 1][0], dictNB[k - 1][0]

        DataSet_Cu.IndexSpotsSet(
            file_to_index,
            key_material,
            emin,
            emax,
            dict_params,
            database,
            use_file=0,
            IMM=False,
            angletol_list=[0.5, 0.5],
            MatchingRate_List=[10, 10, 10],
            nbGrainstoFind=nbGrainstoFind_Cu,
            verbose=0,
            previousResults=previousResults,
        )

        dictspots[k] = DataSet_Cu.getSummaryallData()

        if saveObject:
            dict_spotssetObj[k] = DataSet_Cu

        # filling dicts with indexation results
        for nbgrain in list(range(nbGrainstoFind_Cu)):
            dictMat[k][nbgrain] = DataSet_Cu.dict_grain_matrix[nbgrain]
            dictMR[k][nbgrain] = DataSet_Cu.dict_grain_matching_rate[nbgrain][1]
            dictNB[k][nbgrain] = DataSet_Cu.dict_grain_matching_rate[nbgrain][0]
            dictstrain[k][nbgrain] = DataSet_Cu.dict_grain_devstrain[nbgrain]

        # intermediate saving
        if (k % 10) == 0:
            #            dictRes = dictMat, dictMR, dictNB
            # filepickle = open(os.path.join(dirtowrite, 'dictCu_3g_%04d_%04d' % (nstart, nend)), 'w')
            # pickle.dump(todump, filepickle)
            # filepickle.close()

            with open(
                os.path.join(dirtowrite, "dictCu_3g_%04d_%04d" % (nstart, nend)), "w"
            ) as f:
                pickle.dump(todump, f)

    # filepickle = open(os.path.join(dirtowrite, 'dictCu_3g_%04d_%04d' % (nstart, nend)), 'w')
    # pickle.dump(todump, filepickle)
    # filepickle.close()

    with open(
        os.path.join(dirtowrite, "dictCu_3g_%04d_%04d" % (nstart, nend)), "w"
    ) as f:
        pickle.dump(todump, f)

    #        DataSet_Cu.plotallgrains()

    return todump


#    return DataSet_Si, DataSet_Cu

# ------------  Main            ------------------------------
if __name__ == "__main__":

    res = test_GetProximity2()

    # test_indexation_with_apriori_Matrices()
    #     Test_Common_index_Method()
    # test_indexation_generalRefunitCell_1()
    # test_indexation_generalRefunitCell_2()

    # Test_Common_index_Method()

    """
    # building the map
    # nb elem = 51*23

    list_of_pointsinstance=[]
    list_of_signs=[]
    dic_ID_orient={}
    grain_counter=0
    for m in range(51*23):
        XX=m/51
        YY=m%51
        p=Map_point([XX,YY])
        p.orients=refine_orient(Main_Orient[m],array([3,3,3]))
        list_of_pointsinstance.append(p)

        nb_orient=len(p.orients)
        #print m,nb_orient
        if nb_orient>0:
            point_sign=[]
            for g in range(nb_orient):
                #print g,list(p.orients[g])
                signature_g=from3angles_to_int(list(p.orients[g]))
                point_sign.append(signature_g)
                if signature_g not in dic_ID_orient.values():
                    dic_ID_orient[grain_counter]=signature_g
                    grain_counter+=1
            list_of_signs.append(point_sign)
            p.orients_ID=point_sign
        else:
            list_of_signs.append([-1])
            p.orients_ID=[-1]

    print "I ve found %d grain in the map"%len(dic_ID_orient)
    #adding -1 = None
    dic_ID_orient[-1]=-1
    print dic_ID_orient

    print "Now put the general ID grains in all the point of the map"
    # inverser le dictionnaire...
    def invdict(adict):

        return dict([(val, key) for key, val in adict.items()])

    dic_orient_ID=invdict(dic_ID_orient)

    # lists grain id for each points
    grain_per_point=[map(lambda elem: dic_orient_ID[elem],ls) for ls in list_of_signs]

    # gives for each grain ID, the point instances where they were found ----------------------------
    localisation_grain={}
    for ID in dic_ID_orient.keys():
        localisation_grain[ID]=[]
    for p in list_of_pointsinstance:
        map(lambda cle_ID: localisation_grain[dic_orient_ID[cle_ID]].append(p.pos),p.orients_ID)
    table_pos=[array(localisation_grain[ID]) for ID in dic_ID_orient.keys()]

    """

    # -----------------------------------------------------------------------------------------------
    def make_colormarkerdict():
        """
        build a dictionary of color and marker to distinguish contribution of grains in scatter plot
        """
        # faudrait rendre aleatoire l'indice i_dict pour un meilleur rendu
        colorlist = [ "k", (1.0, 0.0, 0.5),
                    "b", "g", "y", "r", "c", "m", "0.75",
                    (0.2, 0.2, 0.2), (0.0, 0.5, 1.0), ]
        markerlist = ["s", "v", "h", "d", "o", "^", "8"]
        colormarkerdict = {}
        i_dict = 0
        for col in colorlist:
            for mark in markerlist:
                colormarkerdict[i_dict] = [col, mark]
                i_dict = i_dict + 1
        return colormarkerdict

    def plot_mapmajor(listoflocations, xmin, xmax, ymin, ymax):
        """
        example of map plot given list of coordinates ?? ( I do not remember actually)

        """
        import pylab as p

        p.close()
        colormarkerdict = make_colormarkerdict()

        tableofxcoord = []
        tableofycoord = []

        # move the last elem in first position the first grain ID corresponds now to unknown grain
        listoflocations.insert(0, listoflocations.pop(-1))

        for elem in (
            listoflocations[0],
            listoflocations[8],
            listoflocations[16],
            listoflocations[24],
            listoflocations[32],
        ):
            tableofxcoord.append(elem[:, 0])
            tableofycoord.append(elem[:, 1])

        # print "tableofxcoord",tableofxcoord

        p.figure(figsize=(10, 7), dpi=80)

        kk = 0
        for xdata, ydata in zip(tuple(tableofxcoord), tuple(tableofycoord)):
            p.scatter(
                ydata,
                xdata,
                c=colormarkerdict[kk][0],
                marker=colormarkerdict[kk][1],
                s=500,
                alpha=0.5,
            )
            # permutation of x y / array index ?...
            kk = kk + 1

        ticks_x = np.arange(xmin, xmax, 1)
        ticks_y = np.arange(ymin, ymax, 1)
        p.yticks(ticks_x)  # permutation of x y / array index ?...
        p.xticks(ticks_y)
        p.ylim(-1, xmax + 1)
        p.xlim(-1, ymax + 1)
        p.xlabel("x", fontsize=20)
        p.ylabel("y", fontsize=20)
        p.title("UO2_He_103_1275: Grain map. Unknown grains are black squares")
        p.grid(True)
        p.show()

    # plot_mapmajor(table_pos,0,22,0,50)

    # from indexing
    if 0:
        initialparameters = {}

        LaueToolsProjectFolder = os.path.abspath(os.curdir)

        print("LaueToolProjectFolder", LaueToolsProjectFolder)

        MainFolder = os.path.join(LaueToolsProjectFolder, "Examples", "GeGaN")

        print("MainFolder", MainFolder)

        initialparameters["PeakList Folder"] = os.path.join(MainFolder, "datfiles")
        initialparameters["IndexRefine PeakList Folder"] = os.path.join(
            MainFolder, "fitfiles"
        )
        initialparameters["PeakListCor Folder"] = os.path.join(MainFolder, "corfiles")
        initialparameters["PeakList Filename Prefix"] = "nanox2_400_"
        initialparameters["IndexRefine Parameters File"] = os.path.join(
            MainFolder, "GeGaN.irp"
        )
        initialparameters["Detector Calibration File .det"] = os.path.join(
            MainFolder, "calibGe_nanowMARCCD165.det"
        )
        initialparameters["Detector Calibration File (.dat)"] = os.path.join(
            MainFolder, "nanox2_400_0000_LT_1.dat"
        )
        initialparameters["PeakList Filename Suffix"] = ".dat"

        print(initialparameters["IndexRefine Parameters File"])

        dict_param_list = readIndexRefineConfigFile(
            initialparameters["IndexRefine Parameters File"]
        )

        CCDparams, calibmatrix = IOLT.readfile_det(
            initialparameters["Detector Calibration File .det"], nbCCDparameters=8
        )

        Index_Refine_Parameters_dict = {}
        Index_Refine_Parameters_dict["CCDCalibParameters"] = CCDparams[:5]
        Index_Refine_Parameters_dict["pixelsize"] = CCDparams[5]
        Index_Refine_Parameters_dict["framedim"] = CCDparams[6:8]

        Index_Refine_Parameters_dict["PeakList Folder"] = initialparameters[
            "PeakList Folder"
        ]
        Index_Refine_Parameters_dict["PeakListCor Folder"] = initialparameters[
            "PeakListCor Folder"
        ]
        Index_Refine_Parameters_dict["nbdigits"] = 4
        Index_Refine_Parameters_dict["prefixfilename"] = initialparameters[
            "PeakList Filename Prefix"
        ]
        Index_Refine_Parameters_dict["suffixfilename"] = ".dat"
        Index_Refine_Parameters_dict["prefixdictResname"] = (
            initialparameters["PeakList Filename Prefix"] + "_dict_"
        )

        Index_Refine_Parameters_dict["PeakListFit Folder"] = (
            initialparameters["IndexRefine PeakList Folder"] + "_test"
        )
        Index_Refine_Parameters_dict["Results Folder"] = (
            initialparameters["IndexRefine PeakList Folder"] + "_test"
        )

        Index_Refine_Parameters_dict["dict params list"] = dict_param_list
        Index_Refine_Parameters_dict["Spots Order Reference File"] = None

        print(Index_Refine_Parameters_dict)

        index_fileseries_3(
            (0, 1, 1),
            Index_Refine_Parameters_dict=Index_Refine_Parameters_dict,
            saveObject=0,
            verbose=0,
            nb_materials=None,
            build_hdf5=False,
            prefixfortitle="",
            use_previous_results=False,
            CCDCalibdict=None,
        )

    if 1:
        # E1_13_run2_0171.mccd
        Index_Refine_Parameters_dict = {}

        dict_params = [
            {
                "MATCHINGRATE THRESHOLD IAL": 100.0,
                "MATCHINGRATE ANGLE TOL": 0.5,
                "NBMAXPROBED": 30,
                "central spots indices": 5,
                "key material": "ZrO2Y2O3",
                "emin": 5,
                "emax": 23,
                "nbGrainstoFind": 2,
            }
        ]

        CCDcalib = [
            68.61556108571234347,
            940.77744180377760586,
            1005.14714125015211721,
            0.72531401975651599,
            -0.76181063603636234,
        ]

        CCDCalibdict = {}
        CCDCalibdict["CCDCalibParameters"] = CCDcalib
        CCDCalibdict["framedim"] = (2048, 2048)
        CCDCalibdict["detectordiameter"] = 165.0

        Index_Refine_Parameters_dict["CCDCalibParameters"] = CCDcalib
        # , 0.08056640625000000, 2048.00000000000000000, 2048.00000000000000000]
        Index_Refine_Parameters_dict["pixelsize"] = 0.080566
        # 0.08057, 2048, 2048

        Index_Refine_Parameters_dict["dict params list"] = dict_params

        Index_Refine_Parameters_dict[
            "PeakList Folder"
        ] = "/home/micha/LaueTools/Examples/ZrO2"
        Index_Refine_Parameters_dict[
            "PeakListCor Folder"
        ] = "/home/micha/LaueTools/Examples/ZrO2"
        Index_Refine_Parameters_dict["nbdigits"] = 4
        Index_Refine_Parameters_dict["prefixfilename"] = "E1_13_run2_"
        Index_Refine_Parameters_dict["suffixfilename"] = ".dat"
        Index_Refine_Parameters_dict["prefixdictResname"] = "E1_13_run2_dict_"

        Index_Refine_Parameters_dict[
            "PeakListFit Folder"
        ] = "/home/micha/LaueTools/Examples/ZrO2"
        Index_Refine_Parameters_dict[
            "Results Folder"
        ] = "/home/micha/LaueTools/Examples/ZrO2"

        fileindexrange = (171, 177, 3)

        index_fileseries_3(
            fileindexrange,
            Index_Refine_Parameters_dict=Index_Refine_Parameters_dict,
            saveObject=0,
            CCDCalibdict=CCDCalibdict,
        )



r"""
Core module to compute Laue Pattern in various geometry

Main author is J. S. Micha:   micha [at] esrf [dot] fr

version July 2019
from LaueTools package hosted in

http://sourceforge.net/projects/lauetools/

or

https://gitlab.esrf.fr/micha/lauetools
"""
from __future__ import division

import math
import sys

import numpy as np

np.set_printoptions(precision=15)

# LaueTools modules
if sys.version_info.major == 3:
    import builtins
    from . import CrystalParameters as CP
    from . import generaltools as GT
    from . import IOLaueTools as IOLT
    from . dict_LaueTools import (dict_Materials, dict_Extinc, CST_ENERGYKEV,
            DEFAULT_DETECTOR_DISTANCE, DEFAULT_DETECTOR_DIAMETER, DEFAULT_TOP_GEOMETRY)

    from . import LaueGeometry as LTGeo
else:
    import CrystalParameters as CP
    import generaltools as GT
    import IOLaueTools as IOLT
    from dict_LaueTools import (dict_Materials, dict_Extinc, CST_ENERGYKEV,
            DEFAULT_DETECTOR_DISTANCE, DEFAULT_DETECTOR_DIAMETER, DEFAULT_TOP_GEOMETRY)

    # TODO: LTGeo to be removed
    import LaueGeometry as LTGeo

try:
    import generatehkl

    USE_CYTHON = True
except ImportError:
    #print("-- warning. Cython compiled module for fast computation of Laue spots is not installed!")
    USE_CYTHON = False

DEG = np.pi / 180.0

# --- ---------- Spot class
class spot:
    r"""
    Laue Spot class still used...

    .. todo:: To be avoided for greater performance... To be replaced by full numpy computations
    """
    def __init__(self, indice):
        self.Millers = indice
        self.Qxyz = None
        self.EwaldRadius = None
        self.Xcam = None
        self.Ycam = None
        self.Twicetheta = None
        self.Chi = None

    def __hash__(self):
        r"""
        to used indices for defining a key and using set() sorted()  ... object
        comparison
        In our case used for eliminating spots which equal fondamental key
        Warning: Indice must not be changed , intead troubles:
            see http://wiki.python.org/moin/DictionaryKeys
        """
        # listint = tuple(self.have_fond().Millers)
        # still have -2 -4 -2   =  -2 -2 -4
        III = self.have_fond_indices()
        listint = tuple(np.array([100000, 1000, 1]) + III)
        return listint.__hash__()

    def have_fond_indices(self):
        r"""
        return prime indices (fondamental direction)
        """
        interlist = [x for x in list(self.Millers) if x != 0]
        inter = [int(math.fabs(elem)) for elem in interlist]
        # print "interlist ",interlist
        # print "inter ", inter
        return np.array(self.Millers) / GT.pgcdl(inter)

# --- ---------------   PROCEDURES
def Quicklist(OrientMatrix, ReciprocBasisVectors, listRSnorm, lambdamin, verbose=0):
    r"""
    return 6 indices min and max boundary values for each Miller index h, k, l
    to be contained in the largest Ewald Sphere.

    :param OrientMatrix:  orientation matrix (3*3 matrix)
    :param ReciprocBasisVectors:      list of the three vectors a*,b*,c* in the lab frame
                            before rotation with OrientMatrix
    :param listRSnorm:      : list of the three reciprocal space lengthes of a*,b*,c*
    :param lambdamin:       : lambdamin (in Angstrom) corresponding to energy max

    :return: [[hmin,hmax],[kmin,kmax],[lmin,lmax]]
    """
    #     print "OrientMatrix in Quicklist", OrientMatrix

    assert lambdamin > 0, "lambdamin in Quicklist is not positive! %s" % str(lambdamin)

    if isinstance(OrientMatrix, list):
        OrientMatrix = np.array(OrientMatrix)

    if len(OrientMatrix) != 3:
        raise ValueError("Matrix is not 3*3 array !!!%s"%str(OrientMatrix))

    if 0. in listRSnorm or 0 in listRSnorm:
        raise ValueError("listRSnorm contains 0 !! %s"%str(listRSnorm))

    # lengthes of a*,b*,c*
    astarnorm, bstarnorm, cstarnorm = listRSnorm[:3]

    if verbose:
        print("norms of input a*,b*,c* : in Angstrom-1")
        print([astarnorm, bstarnorm, cstarnorm])

    # vec a*,b*,c* input
    vecastar, vecbstar, veccstar = np.array(ReciprocBasisVectors)

    # unit vectors
    # input vec a* normalized , ie vec a* direction
    vecastar_n = vecastar / math.sqrt(sum(vecastar ** 2))
    vecbstar_n = vecbstar / math.sqrt(sum(vecbstar ** 2))
    veccstar_n = veccstar / math.sqrt(sum(veccstar ** 2))

    # vec a* with appropriate length
    astar = vecastar_n * astarnorm
    bstar = vecbstar_n * bstarnorm
    cstar = veccstar_n * cstarnorm

    if verbose:
        print("(a*,b*,c*) in Angstrom-1")
        print(astar)
        print(bstar)
        print(cstar)

    normastar = math.sqrt(sum(astar ** 2))
    normbstar = math.sqrt(sum(bstar ** 2))
    normcstar = math.sqrt(sum(cstar ** 2))

    if verbose:
        print("norms of a*,b*,c* : (must be equal to input norms) in Angstrom-1")
        print(normastar, normbstar, normcstar)

    # rotation matrix or orientation matrix
    matrice_rot = np.array(OrientMatrix)

    if verbose:
        print(matrice_rot)

    rotAstar = np.dot(matrice_rot, astar)  # R(vec a* with appropriate length)
    rotBstar = np.dot(matrice_rot, bstar)
    rotCstar = np.dot(matrice_rot, cstar)

    if verbose:
        print("R(a*, b*, c*) in Angstrom-1")
        print(rotAstar)
        print(rotBstar)
        print(rotCstar)

    # transfer matrix from R(a*),R(b*),R(c*) to X, Y, Z
    matricerotvecstar = np.array([rotAstar, rotBstar, rotCstar]).transpose()
    # transfer matrix from X, Y, Z to R(a*),R(b*),R(c*)
    # inv_matricerotstar = scipy.linalg.basic.inv(matricerotvecstar)
    inv_matricerotstar = np.linalg.inv(matricerotvecstar)

    if verbose:
        print(" ------------------------")
        print("transfer matrix from R(a*),R(b*),R(c*) to X, Y, Z")
        print(matricerotvecstar)

        print(" ------------------------")
        print("transfer matrix from X, Y, Z to R(a*),R(b*),R(c*)")
        print(inv_matricerotstar)

        print("1 / lambdamin:", 1.0 / lambdamin, " Angstrom-1")

    OC = np.array([-1.0 / lambdamin, 0.0, 0.0])  # centre of the largest Ewald sphere in X, Y, Z frame

    OCrot = np.dot(inv_matricerotstar, OC)  # same centre in the R(a*),R(b*),R(c*) frame

    if verbose:
        print("center of largest Ewald Sphere in the R(a*),R(b*),R(c*) frame")
        print(OCrot)

        print("1 / lambdamin in corresponding R(a*),R(b*),R(c*) units")
        print([1.0 / lambdamin / normastar,
                1.0 / lambdamin / normbstar,
                1.0 / lambdamin / normcstar])

    # for non alpha = beta = gamma = 90 deg case
    # Calculate the crossproduct of rotated vector to correct properly the range on each rotated vector

    crossastar = np.cross(rotBstar, rotCstar) / math.sqrt(sum(np.cross(rotBstar, rotCstar) ** 2))  # normalised cross vector for reference
    crossbstar = np.cross(rotCstar, rotAstar) / math.sqrt(sum(np.cross(rotCstar, rotAstar) ** 2))
    crosscstar = np.cross(rotAstar, rotBstar) / math.sqrt(sum(np.cross(rotAstar, rotBstar) ** 2))

    if verbose:
        print("crossedvector", crossastar, crossbstar, crosscstar)

    cosanglea = np.dot(rotAstar / normastar, crossastar)
    cosangleb = np.dot(rotBstar / normbstar, crossbstar)
    cosanglec = np.dot(rotCstar / normcstar, crosscstar)

    if verbose:
        print("cosangle", [cosanglea, cosangleb, cosanglec])

    hmin = GT.properinteger(OCrot[0] - 1.0 / lambdamin / normastar / cosanglea)
    kmin = GT.properinteger(OCrot[1] - 1.0 / lambdamin / normbstar / cosangleb)
    lmin = GT.properinteger(OCrot[2] - 1.0 / lambdamin / normcstar / cosanglec)

    hmax = GT.properinteger(OCrot[0] + 1.0 / lambdamin / normastar / cosanglea)
    kmax = GT.properinteger(OCrot[1] + 1.0 / lambdamin / normbstar / cosangleb)
    lmax = GT.properinteger(OCrot[2] + 1.0 / lambdamin / normcstar / cosanglec)

    # SECURITY -1
    Hmin = min(hmin, hmax) - 1
    Kmin = min(kmin, kmax) - 1
    Lmin = min(lmin, lmax) - 1

    # SECURITY +1 and +1 because of exclusion convention for slicing in python
    # (superior limit in range is excluded)
    Hmax = max(hmin, hmax) + 2
    Kmax = max(kmin, kmax) + 2
    Lmax = max(lmin, lmax) + 2

    try:
        list_hkl_limits = [[int(Hmin), int(Hmax)],
                        [int(Kmin), int(Kmax)],
                        [int(Lmin), int(Lmax)]]
        return list_hkl_limits
    except ValueError:
        return None


def genHKL_np(listn, Extinc):
    r"""
    Generate all Miller indices hkl from indices limits given  by listn
    and taking into account for systematic exctinctions

    :param listn: Miller indices limits (warning: these lists are used in python range (last index is excluded))
    :type listn: [[hmin,hmax],[kmin,kmax],[lmin,lmax]]

    :param Extinc: label corresponding to systematic exctinction
        rules on h k and l miller indics such as ('fcc', 'bcc', 'dia', ...) or 'no' for any rules
    :type Extinc: string

    :return: array of [h,k,l]

    .. note:: node [0,0,0] is excluded
    """
    if listn is None:
        raise ValueError("hkl ranges are undefined")

    #    print "inside genHKL_np"
    if isinstance(listn, list) and len(listn) == 3:
        try:
            n_h_min, n_h_max = listn[0]
            n_k_min, n_k_max = listn[1]
            n_l_min, n_l_max = listn[2]
        except:
            raise TypeError("arg #1 has not the shape (3, 2)")
    else:
        raise TypeError("arg #1 is not a list or has not 3 elements")

    nbelements = (n_h_max - n_h_min) * (n_k_max - n_k_min) * (n_l_max - n_l_min)

    assert nbelements > 0

    if (not isinstance(nbelements, int)) or nbelements <= 0.0:
        raise ValueError("Needs (3,2) list of sorted integers")

    if Extinc not in list(dict_Extinc.values()):
        raise ValueError('Could not understand extinction code: " %s " ' % str(Extinc))

    HKLraw = np.mgrid[n_h_min:n_h_max, n_k_min:n_k_max, n_l_min:n_l_max]
    HKLs = HKLraw.shape
    #    print "HKLs", HKLs
    HKL = HKLraw.T.reshape((HKLs[1] * HKLs[2] * HKLs[3], HKLs[0]))

    return CP.ApplyExtinctionrules(HKL, Extinc)


# --- -----------------------  Main procedures
def parse_grainparameters(SingleCrystalParams):
    r"""
    extract and check type of the 4 elements of SingleCrystalParams

    .. todo:: to be used
    """
    try:
        Bmatrix, Extinc, Orientmatrix, key_for_dict = SingleCrystalParams
    except IndexError:
        raise IndexError("List of input parameters for crystal Laue pattern simulation "
                            "has %d elements instead of 4 !" % len(SingleCrystalParams))

    if Extinc not in list(dict_Extinc.values()):
        raise ValueError("wrong code %s for extinction !" % Extinc)
    if np.array(Bmatrix).shape != (3, 3):
        raise ValueError("wrong B matrix format!")
    if np.array(Orientmatrix).shape != (3, 3):
        raise ValueError("wrong Orientmatrix format!")
    return Bmatrix, Extinc, Orientmatrix, key_for_dict


def getLaueSpots(wavelmin, wavelmax, crystalsParams, kf_direction=DEFAULT_TOP_GEOMETRY,
                                            OpeningAngleCollection=22.0,
                                            fastcompute=0,
                                            ResolutionAngstrom=False,
                                            verbose=1,
                                            dictmaterials=None):
    r"""
    Compute Qxyz vectors and corresponding HKL miller indices for nodes in recicprocal space that can be measured
    for the given detection geometry and energy bandpass configuration.

    :param wavelmin:   smallest wavelength in Angstrom
    :param wavelmax:  largest wavelength in Angstrom

    :param crystalsParams: list of *SingleCrystalParams*, each of them being a list
        of 4 elements for crystal orientation and strain properties:

        * [0](array): is the B matrix a*,b*,c* vectors are expressed in column
            in LaueTools frame in reciprocal angstrom units

        * [1](str): peak Extinction rules ('no','fcc','dia', etc...)

        * [2](array): orientation matrix

        * [3](str): key for material element

    :param kf_direction: string defining the average geometry, mean value of exit scattered vector:
        'Z>0'   top spots

        'Y>0'   one side spots (towards hutch door)

        'Y<0'   other side spots

        'X>0'   transmission spots

        'X<0'   backreflection spots
    :param fastcompute:
        * 1, compute reciprocal space (RS) vector BUT NOT the Miller indices

        * 0, returns both RS vectors (normalised) and Miller indices

    :param ResolutionAngstrom:
        * scalar, smallest interplanar distance ordered in crystal in angstrom.

        * None, all reflections will be calculated that can be time-consuming for large unit cell

    :return:
        * list of [Qx,Qy,Qz]s for each grain, list of [H,K,L]s for each grain (fastcompute = 0)

        * list of [Qx,Qy,Qz]s for each grain, None  (fastcompute = 1)

    .. caution::
        This method doesn't create spot instances.

        This is done in filterLaueSpots with fastcompute = 0
    .. caution::
        finer selection of nodes : on camera , without harmonics can be
        done later with filterLaueSpots()

    .. note:: lauetools laboratory frame is in this case:
        x// ki (center of ewald sphere has negative x component)
        z perp to x and belonging to the plane defined by x and dd vectors
        (where dd vector is the smallest vector joining sample impact point and points on CCD plane)
        y is perpendicular to x and z

    """
    if isinstance(wavelmin, (float, int)) and isinstance(wavelmax, (float, int)):
        if wavelmin < wavelmax and wavelmin > 0:
            wlm = wavelmin * 1.0
            wlM = wavelmax * 1.0
        else:
            raise ValueError("wavelengthes must be positive and ordered")
    else:
        raise ValueError("wavelengthes must have numerical values")

    if not isinstance(crystalsParams, list):
        raise ValueError("Grains parameters list is not correct. It Must be: [grain] for 1 grain or [grain1, grain2] for 2 grains...")

    nb_of_grains = len(crystalsParams)

    if verbose:
        print("energy range: %.2f - %.2f keV" % (CST_ENERGYKEV / wlM, CST_ENERGYKEV / wlm))
        print("Number of grains: ", nb_of_grains)

    if dictmaterials is None:
        dictmaterials = dict_Materials

    # loop over grains
    for i in list(range(nb_of_grains)):
        try:
            key_material = dictmaterials[crystalsParams[i][3]][0]
        except (IndexError, TypeError, KeyError):
            smsg = "wrong type of input paramters: must be a list of 4 elements"
            raise ValueError(smsg)
        if verbose:
            print("# grain:  ", i, " made of ", key_material)
            print(" crystal parameters:", crystalsParams[i])

    wholelistvecfiltered = []
    wholelistindicesfiltered = []

    # calculation of RS lattice nodes in lauetools laboratory frame and indices

    # loop over grains
    for i in list(range(nb_of_grains)):

        (Bmatrix, Extinc,
        Orientmatrix, key_for_dict) = parse_grainparameters(crystalsParams[i])

        if verbose:
            print("\nin getLaueSpots()")
            print("Bmatrix", Bmatrix)
            print("Orientmatrix", Orientmatrix)
            print("key_for_dict", key_for_dict)
            print("Extinc", Extinc)
            print("")

        Bmatrix = np.array(Bmatrix)

        # generation of hkl nodes from Bmatrix

        listvecstarlength = np.sqrt(np.sum(Bmatrix ** 2, axis=0))

        # print "listvecstarlength",listvecstarlength

        # in Bmatrix.T,  a*, b* ,c* are rows of this argument
        #  B matrix in q= Orientmatrix B G formula
        # limitation of probed h k and l ranges
        list_hkl_limits = Quicklist(Orientmatrix, Bmatrix.T, listvecstarlength, wlm, verbose=0)

        # Loop over h k l
        # ----cython optimization
        global USE_CYTHON
        if USE_CYTHON:
            hlim, klim, llim = list_hkl_limits
            hmin, hmax = hlim
            kmin, kmax = klim
            lmin, lmax = llim
            dict_extinc = {"no": 0, "fcc": 1, "dia": 2}
            try:
                ExtinctionCode = dict_extinc[Extinc]
                SPECIAL_EXTINC = False
            except KeyError:
                ExtinctionCode = 0
                SPECIAL_EXTINC = True

            #  print "\n\n*******\nUsing Cython optimization\n********\n\n"
            hkls, counter = generatehkl.genHKL(hmin, hmax, kmin, kmax, lmin, lmax, ExtinctionCode)

            table_vec = hkls[:counter]

            # TODO need to remove element [0,0,0] or naturally removed in ?
            if SPECIAL_EXTINC:
                #  print "special extinction"
                table_vec = CP.ApplyExtinctionrules(table_vec, Extinc)

        if not USE_CYTHON:
            table_vec = genHKL_np(list_hkl_limits, Extinc)

        Orientmatrix = np.array(Orientmatrix)

        listrotvec = np.dot(Orientmatrix, np.dot(Bmatrix, table_vec.T))

        listrotvec_X = listrotvec[0]
        listrotvec_Y = listrotvec[1]
        listrotvec_Z = listrotvec[2]

        arraysquare = listrotvec_X ** 2 + listrotvec_Y ** 2 + listrotvec_Z ** 2

        # removing all spots that have positive X component
        cond_Xnegativeonly = listrotvec_X < 0.
        listrotvec_X = listrotvec_X[cond_Xnegativeonly]
        listrotvec_Y = listrotvec_Y[cond_Xnegativeonly]
        listrotvec_Z = listrotvec_Z[cond_Xnegativeonly]
        arraysquare = arraysquare[cond_Xnegativeonly]

        table_vec = table_vec[cond_Xnegativeonly]

        if sys.version_info.major == 3:
            listObj = builtins.list
        else:
            listObj = list

        # Kf direction selection
        # top reflection 2theta = 90
        if kf_direction == "Z>0":
            KF_condit = listrotvec_Z > 0.0
        # side reflection  2theta = 90
        elif kf_direction == "Y>0":
            KF_condit = listrotvec_Y > 0
        # side reflection  2theta = 90
        elif kf_direction == "Y<0":
            KF_condit = listrotvec_Y < 0

        # x > -R transmission 2theta = 0
        elif kf_direction == "X>0":
            KF_condit = (listrotvec_X + 1.0 / (2.0 * np.abs(listrotvec_X) / arraysquare) > 0)

        # x < -R back reflection 2theta = 180
        elif kf_direction == "X<0":
            KF_condit = (listrotvec_X + 1.0 / (2.0 * np.abs(listrotvec_X) / arraysquare) < 0)
        #   all spots inside the two ewald's sphere with scattered beams in any direction'
        elif kf_direction == "4PI":
            KF_condit = np.ones_like(listrotvec_X) * True
        # user's definition of mean kf vector
        # [2theta , chi]= / kf = [cos2theta,sin2theta*sinchi,sin2theta*coschi]
        elif isinstance(kf_direction, (listObj, np.array)):
            print("\nUSING user defined LauePattern Region\n")
            if len(kf_direction) != 2:
                raise ValueError("kf_direction must be defined by a list of two angles !")
            else:
                kf_2theta, kf_chi = kf_direction
                kf_2theta, kf_chi = kf_2theta * DEG, kf_chi * DEG

                qmean_theta, qmean_chi = kf_2theta / 2.0, kf_chi

                print("central q angles", qmean_theta, qmean_chi)

                # q.qmean >0
                #    KF_condit = (-listrotvec_X * np.sin(qmean_theta) + \
                #                 listrotvec_Y * np.cos(qmean_theta) * np.sin(qmean_chi) + \
                #                 listrotvec_Z * np.cos(qmean_theta) * np.cos(qmean_chi)) > 0

                #    # angle(q, qmean) < 45 deg
                #     AngleMax = OpeningAngleCollection * DEG
                #   KF_condit = np.arccos(((-listrotvec_X * np.sin(qmean_theta) + \
                #                            listrotvec_Y * np.cos(qmean_theta) * np.sin(qmean_chi) + \
                #                            listrotvec_Z * np.cos(qmean_theta) * np.cos(qmean_chi))) / np.sqrt(arraysquare)) < AngleMax

                # angle(kf, kfmean) < 45 deg
                AngleMax = OpeningAngleCollection * DEG

                Rewald = arraysquare / (2 * np.fabs(listrotvec_X))
                kfsquare = Rewald ** 2
                KF_condit = (np.arccos((
                            (listrotvec_X + Rewald) * np.cos(kf_2theta)
                            + listrotvec_Y * np.sin(kf_2theta) * np.sin(kf_chi)
                            + listrotvec_Z * np.sin(kf_2theta) * np.cos(kf_chi)
                        ) / np.sqrt(kfsquare)) < AngleMax)
        else:
            raise ValueError("kf_direction '%s' is not understood!" % str(kf_direction))

        #        print "nb of KF_condit", len(np.where(KF_condit == True)[0])

        # vectors inside the two Ewald's spheres
        # last condition could be perhaps put before on
        # listrotvec_X[listrotvec_Z > 0]...
        Condit = np.logical_and(np.logical_and(
                                        ((listrotvec_X * 2.0 / wlm + arraysquare) <= 0.0),
                                        ((listrotvec_X * 2.0 / wlM + arraysquare) > 0.0),
                                    ), (KF_condit))
        #   print "nb of spots in spheres and in towards CCD region", len(np.where(Condit == True)[0])

        if fastcompute == 0:
            #             print 'Using detailled computation mode'

            if ResolutionAngstrom:
                # #crude resolution limitation 2Angstrom for lattice [20 4.8 49]
                # Hcond = np.abs(table_vec[:,0])<10
                # Kcond = np.abs(table_vec[:,1])<3
                # Lcond = np.abs(table_vec[:,2])<25
                # ConditResolution = np.logical_and(np.logical_and((Hcond) ,
                #                                                  (Kcond)),
                #                                                   (Lcond))
                ConditResolution = arraysquare < (1.0 / ResolutionAngstrom) ** 2

                Condit = np.logical_and(Condit, ConditResolution)

            # print "Condit",Condit
            fil_X = np.compress(Condit, listrotvec_X)
            fil_Y = np.compress(Condit, listrotvec_Y)
            fil_Z = np.compress(Condit, listrotvec_Z)

            # print "fil_Z",fil_Z
            fil_H = np.compress(Condit, table_vec[:, 0])
            fil_K = np.compress(Condit, table_vec[:, 1])
            fil_L = np.compress(Condit, table_vec[:, 2])

            listvecfiltered = np.transpose(np.array([fil_X, fil_Y, fil_Z]))
            listindicesfiltered = np.transpose(np.array([fil_H, fil_K, fil_L]))

            #  print "listindicesfiltered", len(listindicesfiltered)

            # if 0:
            # print "listrotvec[0]",listrotvec[0]
            # print "fgfghfg",fil_X
            # Energyarray = CST_ENERGYKEV*2*np.abs(listrotvec_X) / arraysquare
            # Efiltered = np.compress(Condit, Energyarray)
            # print "energyfiltered",Efiltered
            # print "listindicesfiltered",listindicesfiltered

            # to return
            wholelistvecfiltered.append(listvecfiltered)
            wholelistindicesfiltered.append(listindicesfiltered)

            if verbose:
                print("Orientation matrix", Orientmatrix)
                print("# grain: ", i)
                print("Number of spots for # grain ", i, ": ", len(listvecfiltered))

        # only q vectors are calculated, not Miller indices
        elif fastcompute == 1:

            if ResolutionAngstrom:

                # #crude resolution limitation 2Angstrom for lattice [20 4.8 49]
                # Hcond = np.abs(table_vec[:,0])<10
                # Kcond = np.abs(table_vec[:,1])<3
                # Lcond = np.abs(table_vec[:,2])<25
                # ConditResolution = np.logical_and(np.logical_and((Hcond) ,
                #                                                  (Kcond)),
                #                                                   (Lcond))

                ConditResolution = arraysquare < (1.0 / ResolutionAngstrom) ** 2

                Condit = np.logical_and(Condit, ConditResolution)

            fil_X = np.compress(Condit, listrotvec_X)
            fil_Y = np.compress(Condit, listrotvec_Y)
            fil_Z = np.compress(Condit, listrotvec_Z)

            # to return
            listvecfiltered = np.transpose(np.array([fil_X, fil_Y, fil_Z]))
            wholelistvecfiltered.append(listvecfiltered)
            # and wholelistindicesfiltered (hkl) is empty

            if verbose:
                print("Orientation matrix", Orientmatrix)
                print("# grain: ", i)
                print("Rotating all spots")
                print("Nb of spots for # grain  ", i, ": ", len(listvecfiltered))

    return wholelistvecfiltered, wholelistindicesfiltered


def create_spot(pos_vec, miller, detectordistance, allattributes=False, pixelsize=165.0 / 2048,
                                                                                dim=(2048, 2048)):
    r""" From reciprocal space position and 3 miller indices
    create a spot instance (on top camera geometry)

    :param pos_vec: 3D vector
    :type pos_vec: list of 3 float
    :param  miller: list of 3 miller indices
    :param detectordistance: approximate distance detector sample (to compute complementary spots attributes)
    :param  allattributes: False or 0  not to compute complementary spot attributes
    :param  allattributes: boolean

    :return: spot instance

    .. note:: spot.Qxyz is a vector expressed in lauetools frame

    X along x-ray and Z towards CCD when CCD on top and y towards experimental hutch door
    """
    spotty = spot(miller)
    spotty.Qxyz = pos_vec
    vecres = np.array(pos_vec)
    spotty.EwaldRadius = np.dot(vecres, vecres) / (2.0 * math.fabs(vecres[0]))
    if spotty.Qxyz[2] > 0:
        normkout = math.sqrt((spotty.Qxyz[0] + spotty.EwaldRadius) ** 2
            + spotty.Qxyz[1] ** 2
            + spotty.Qxyz[2] ** 2)

        if allattributes not in (False, 0):
            X = (detectordistance
                * (spotty.Qxyz[0] + spotty.EwaldRadius)
                / spotty.Qxyz[2])
            Y = detectordistance * (spotty.Qxyz[1]) / spotty.Qxyz[2]
            spotty.Xcam = -X / pixelsize + dim[0] / 2
            spotty.Ycam = -Y / pixelsize + dim[1] / 2

        spotty.Twicetheta = (math.acos((spotty.Qxyz[0] + spotty.EwaldRadius) / normkout) / DEG)
        #        spotty.Chi = math.atan(spotty.Qxyz[1] * 1. / spotty.Qxyz[2]) / DEG
        spotty.Chi = math.atan2(spotty.Qxyz[1] * 1.0, spotty.Qxyz[2]) / DEG

    return spotty


def create_spot_np(Qxyz, miller, detectordistance, allattributes=False,
                                                    pixelsize=165.0 / 2048, dim=(2048, 2048)):
    r""" From reciprocal space position and 3 miller indices
    create a spot instance (on top camera geometry)

    :param pos_vec: 3D vector
    :type pos_vec: list of 3 float
    :param  miller: list of 3 miller indices
    :param detectordistance: approximate distance detector sample (to compute complementary spots attributes)
    :param  allattributes: False or 0  not to compute complementary spot attributes
    :param  allattributes: boolean

    :return: spot instance

    .. note:: spot.Qxyz is a vector expressed in lauetools frame

    X along x-ray and Z towards CCD when CCD on top and y towards experimental hutch door
    """
    #     print "Qxyz", Qxyz
    EwaldRadius = np.sum(Qxyz ** 2, axis=1) / (2.0 * np.abs(Qxyz[:, 0]))
    ki = np.zeros((len(Qxyz), 3))
    ki[:, 0] = EwaldRadius

    kout = ki + Qxyz
    #     if spotty.Qxyz[2] > 0:
    normkout = np.sqrt(np.sum(kout ** 2, axis=1))

    Twicetheta = np.arccos((kout[:, 0]) / normkout) / DEG
    Chi = np.arctan2(kout[:, 1] * 1.0, kout[:, 2]) / DEG

    Energy = EwaldRadius * CST_ENERGYKEV

    return Twicetheta, Chi, Energy, miller


def create_spot_4pi(pos_vec, miller, detectordistance, allattributes=0, pixelsize=165.0 / 2048,
                                                                                dim=(2048, 2048)):
    r""" From reciprocal space position and 3 miller indices
    create a spot scattered in 4pi steradian no camera position

    .. note:: spot.Qxyz is a vector expressed in lauetools frame

    X along x-ray and Z towards CCD when CCD on top and y towards experimental hutch door
    """
    spotty = spot(miller)
    spotty.Qxyz = pos_vec
    vecres = np.array(pos_vec)
    spotty.EwaldRadius = np.dot(vecres, vecres) / (2.0 * math.fabs(vecres[0]))

    normkout = math.sqrt((spotty.Qxyz[0] + spotty.EwaldRadius) ** 2
        + spotty.Qxyz[1] ** 2
        + spotty.Qxyz[2] ** 2)

    spotty.Twicetheta = (math.acos((spotty.Qxyz[0] + spotty.EwaldRadius) / normkout) / DEG)
    #    spotty.Chi = math.atan(spotty.Qxyz[1] * 1. / spotty.Qxyz[2]) / DEG
    spotty.Chi = math.atan2(spotty.Qxyz[1] * 1.0, spotty.Qxyz[2]) / DEG

    return spotty


def create_spot_side_pos(pos_vec, miller, detectordistance,
                                        allattributes=0,
                                        pixelsize=165.0 / 2048,
                                        dim=(2048, 2048)):
    r""" From reciprocal space position and 3 miller indices
    create a spot on side camera
    """
    spotty = spot(miller)
    spotty.Qxyz = pos_vec
    vecres = np.array(pos_vec)
    spotty.EwaldRadius = np.dot(vecres, vecres) / (2.0 * math.fabs(vecres[0]))
    if spotty.Qxyz[1] > 0:
        normkout = math.sqrt(
            (spotty.Qxyz[0] + spotty.EwaldRadius) ** 2
            + spotty.Qxyz[1] ** 2
            + spotty.Qxyz[2] ** 2)

        if not allattributes:
            X = (detectordistance * (spotty.Qxyz[0] + spotty.EwaldRadius) / spotty.Qxyz[1])
            Y = detectordistance * (spotty.Qxyz[2]) / spotty.Qxyz[1]
            #             spotty.Xcam = X / pixelsize + dim[0] / 2
            #             spotty.Ycam = Y / pixelsize + dim[1] / 2
            spotty.Xcam = X / pixelsize
            spotty.Ycam = Y / pixelsize

        spotty.Twicetheta = (math.acos((spotty.Qxyz[0] + spotty.EwaldRadius) / normkout) / DEG)
        # spotty.Chi = math.atan(spotty.Qxyz[1]*1. / spotty.Qxyz[2])/DEG
        spotty.Chi = math.atan2(spotty.Qxyz[1] * 1.0, spotty.Qxyz[2]) / DEG

    return spotty


def create_spot_side_neg(pos_vec, miller, detectordistance, allattributes=0, pixelsize=165.0 / 2048,
                                            dim=(2048, 2048)):
    r""" From reciprocal space position and 3 miller indices
    create a spot on neg side camera

    .. todo:: Update with dim as other create_spot()
    """
    spotty = spot(miller)
    spotty.Qxyz = pos_vec
    vecres = np.array(pos_vec)
    spotty.EwaldRadius = np.dot(vecres, vecres) / (2.0 * math.fabs(vecres[0]))

    if spotty.Qxyz[1] < 0.0:
        normkout = math.sqrt(
            (spotty.Qxyz[0] + spotty.EwaldRadius) ** 2
            + spotty.Qxyz[1] ** 2
            + spotty.Qxyz[2] ** 2)

        if not allattributes:
            X = (-detectordistance * (spotty.Qxyz[0] + spotty.EwaldRadius) / spotty.Qxyz[1])
            Y = detectordistance * (spotty.Qxyz[2]) / spotty.Qxyz[1]
            spotty.Xcam = X / pixelsize + dim[0] / 2.0
            spotty.Ycam = Y / pixelsize + dim[1] / 2.0

        spotty.Twicetheta = (math.acos((spotty.Qxyz[0] + spotty.EwaldRadius) / normkout) / DEG)
        # spotty.Chi = math.atan(spotty.Qxyz[1]*1. / spotty.Qxyz[2])/DEG
        spotty.Chi = math.atan2(spotty.Qxyz[1] * 1.0, spotty.Qxyz[2]) / DEG

    return spotty


def create_spot_front(pos_vec, miller, detectordistance, allattributes=0,
                                                        pixelsize=165.0 / 2048,
                                                        dim=(2048, 2048)):
    r""" From reciprocal space position and 3 miller indices
    create a spot on forward direction transmission geometry
    """
    #     print "use create_spot_front"
    spotty = spot(miller)
    spotty.Qxyz = pos_vec
    vecres = np.array(pos_vec)
    spotty.EwaldRadius = np.dot(vecres, vecres) / (2.0 * math.fabs(vecres[0]))

    if miller[0] == 0 and miller[1] == 0 and miller[2] == 0:
        # print("MILLER == MILER ==0")
        return None

    if spotty.Qxyz[0] < 0.0:
        #print('good reflection for transmission')
        abskx = math.fabs(spotty.Qxyz[0] + spotty.EwaldRadius)
        normkout = 1.0 * math.sqrt(abskx ** 2 + spotty.Qxyz[1] ** 2 + spotty.Qxyz[2] ** 2)

        # if qx < -R   no spot in transmission mode (qx is <0)
        if spotty.Qxyz[0] + spotty.EwaldRadius <= 0.:
            return None

        X = -detectordistance * (spotty.Qxyz[1]) / abskx
        Y = -detectordistance * (spotty.Qxyz[2]) / abskx
        spotty.Xcam = X / pixelsize + dim[0] / 2.0
        spotty.Ycam = Y / pixelsize + dim[1] / 2.0

        spotty.Twicetheta = (math.acos((spotty.Qxyz[0] + spotty.EwaldRadius) / normkout) / DEG)
        # spotty.Chi = math.atan(spotty.Qxyz[1]*1. / spotty.Qxyz[2])/DEG
        spotty.Chi = math.atan2(spotty.Qxyz[1] * 1.0, spotty.Qxyz[2]) / DEG

        return spotty

    else:
        return None


def create_spot_back(pos_vec, miller, detectordistance, allattributes=0,
                                                        pixelsize=165.0 / 2048, dim=(2048, 2048)):
    r""" From reciprocal space position and 3 miller indices
    create a spot on backward directions i.e.  back reflection geometry
    """
    spotty = spot(miller)
    spotty.Qxyz = pos_vec
    vecres = np.array(pos_vec)
    spotty.EwaldRadius = np.dot(vecres, vecres) / (2.0 * math.fabs(vecres[0]))
    if spotty.Qxyz[0] < 0.0:
        abskx = math.fabs(spotty.Qxyz[0] + spotty.EwaldRadius)
        normkout = math.sqrt(abskx ** 2 + spotty.Qxyz[1] ** 2 + spotty.Qxyz[2] ** 2)

        if not allattributes:
            X = detectordistance * (spotty.Qxyz[1]) / abskx
            Y = detectordistance * (spotty.Qxyz[2]) / abskx
            spotty.Xcam = X / pixelsize + dim[0] / 2.0
            spotty.Ycam = Y / pixelsize + dim[1] / 2.0

        spotty.Twicetheta = (math.acos((spotty.Qxyz[0] + spotty.EwaldRadius) / normkout) / DEG)
        # spotty.Chi = math.atan(spotty.Qxyz[1]*1. / spotty.Qxyz[2])/DEG
        spotty.Chi = math.atan2(spotty.Qxyz[1] * 1.0, spotty.Qxyz[2]) / DEG

    return spotty

def filterQandHKLvectors(vec_and_indices, detectordistance, detectordiameter, kf_direction='Z>0'):
    """filter vector Q and HKL in vec_and_indices

    :param vec_and_indices: arrays of vectors Q and HKL
    :type vec_and_indices: [Qs, HKLs]
    :param detectordistance: distance detector sample (mm)
    :type detectordistance: float
    :param detectordiameter: detector diameter (mm)
    :type detectordiameter: float
    :param kf_direction: geometry of detection label or two angles giving 2theta chi direction of detector, defaults to 'Z>0'
    :type kf_direction: str or 2 floats, optional
    :raises ValueError: [description]
    :raises ValueError: [description]
    :return: [oncam_vec], [oncam_HKL]
    :rtype: [np.array of 3 elements, np.array of 3 elements]
    """
    Qvectors_list, HKLs_list = vec_and_indices
    Qx = Qvectors_list[0][:, 0] * 1.0
    Qy = Qvectors_list[0][:, 1] * 1.0
    Qz = Qvectors_list[0][:, 2] * 1.0

    indi_H = HKLs_list[0][:, 0]
    indi_K = HKLs_list[0][:, 1]
    indi_L = HKLs_list[0][:, 2]

    Qsquare = Qx ** 2 + Qy ** 2 + Qz ** 2

    # (proportional to photons Energy)
    Rewald = Qsquare / 2.0 / np.abs(Qx)

    # Kf direction selection
    if kf_direction == "Z>0":  # top reflection geometry
        ratiod = detectordistance / Qz
        Ycam = ratiod * (Qx + Rewald)
        Xcam = ratiod * (Qy)
    elif kf_direction == "Y>0":  # side reflection geometry (for detector between the GMT hutch door and the sample (beam coming from right to left)
        ratiod = detectordistance / Qy
        Xcam = ratiod * (Qx + Rewald)
        Ycam = ratiod * (Qz)
    elif kf_direction == "Y<0":  # other side reflection
        ratiod = detectordistance / np.abs(Qy)
        Xcam = -ratiod * (Qx + Rewald)
        Ycam = ratiod * (Qz)
    elif kf_direction == "X>0":  # transmission geometry
        ratiod = detectordistance / np.abs(Qx + Rewald)
        Xcam = -1.0 * ratiod * (Qy)
        Ycam = -ratiod * (Qz)
    elif kf_direction == "X<0":  # back reflection geometry
        ratiod = detectordistance / np.abs(Qx + Rewald)
        Xcam = ratiod * (Qy)
        Ycam = ratiod * (Qz)
    elif kf_direction == "4PI":  # to keep all scattered spots
        Xcam = np.zeros_like(Qy)
        Ycam = np.zeros_like(Qy)

    elif isinstance(kf_direction, (list, np.array)):
        if len(kf_direction) != 2:
            raise ValueError(
                "kf_direction must be defined by a list of two angles !")
        else:
            Xcam = np.zeros_like(Qy)
            Ycam = np.zeros_like(Qy)
    else:
        raise ValueError("Unknown laue geometry code for kf_direction parameter")

    # print "Xcam, Ycam",Xcam
    # print Ycam
    # print "******************"
    # On camera filter
    # print "detectordiameter", detectordiameter
    halfCamdiametersquare = (detectordiameter / 2.0) ** 2
    # TODO: should contain Xcam-Xcamcen (mm) and Ycam-Ycamcen with Xcamcen, Ycamcen
    # given by user
    onCam_cond = Xcam ** 2 + Ycam ** 2 <= halfCamdiametersquare
    # resulting arrays
    oncam_Qx = np.compress(onCam_cond, Qx)
    oncam_Qy = np.compress(onCam_cond, Qy)
    oncam_Qz = np.compress(onCam_cond, Qz)

    oncam_vec = np.array([oncam_Qx, oncam_Qy, oncam_Qz]).T

    oncam_H = np.compress(onCam_cond, indi_H)
    oncam_K = np.compress(onCam_cond, indi_K)
    oncam_L = np.compress(onCam_cond, indi_L)
    oncam_HKL = np.transpose(np.array([oncam_H, oncam_K, oncam_L]))

    return [oncam_vec], [oncam_HKL]


def filterLaueSpots(vec_and_indices, HarmonicsRemoval=1,
                                    fastcompute=0,
                                    kf_direction=DEFAULT_TOP_GEOMETRY,
                                    fileOK=0,
                                    detectordistance=DEFAULT_DETECTOR_DISTANCE,
                                    detectordiameter=DEFAULT_DETECTOR_DIAMETER,
                                    pixelsize=165.0 / 2048,
                                    dim=(2048, 2048),
                                    linestowrite=[[""]],
                                    verbose=0):
    r""" Calculates list of grains spots on camera and without harmonics
    and on CCD camera from [[spots grain 0],[spots grain 1],etc] =>
    returns [[spots grain 0],[spots grain 1],etc] w / o harmonics and on camera  CCD

    :param vec_and_indices: list of elements corresponding to 1 grain, each element is composed by
        * [0] array of vector

        * [1] array of indices

    :param HarmonicsRemoval: 1, removes harmonics according to their miller indices
        (only for fastcompute = 0)

    :param fastcompute:
        * 1, outputs a list for each grain of 2theta spots and a list for each grain of chi spots
            (HARMONICS spots are still HERE!)
        * 0, outputs list for each grain of spots with

    :param kf_direction: label for detection geometry (CCD plane with respect to the incoming beam and sample)
    :type kf_direction: string

    :return:
        * list of spot instances if fastcompute=0

        * 2theta, chi          if fastcompute=1

    .. note::
        * USED IMPORTANTLY in lauecore.SimulateResults  lauecore.SimulateLaue
        * USED in matchingrate.AngularResidues
        * USED in ParametricLaueSimulator.dosimulation_parametric
        * USED in AutoindexationGUI.OnSimulate_S3, DetectorCalibration.Reckon_2pts, and others

    .. todo::
        add dim in create_spot in various geometries
    """
    # print("filterLaueSpots !!!!!")
    try:
        Qvectors_list, HKLs_list = vec_and_indices
    except ValueError:
        raise ValueError("vec_and_indices has not two elements!")

    try:
        nbofgrains = len(Qvectors_list)
    except TypeError:
        raise TypeError("Qvectors_list is not a iterable !")

    if not isinstance(Qvectors_list[0], np.ndarray):
        raise TypeError("first element of Qvectors_list is not surprisingly an array")

    # preparing list of results
    if fastcompute == 0:
        ListSpots_Oncam_wo_harmonics = emptylists(nbofgrains)
    elif fastcompute == 1:
        Oncam2theta = emptylists(nbofgrains)
        Oncamchi = emptylists(nbofgrains)

    # loop over grains
    totalnbspots = 0
    for grainindex in list(range(nbofgrains)):
        try:
            Qx = Qvectors_list[grainindex][:, 0] * 1.0
            Qy = Qvectors_list[grainindex][:, 1] * 1.0
            Qz = Qvectors_list[grainindex][:, 2] * 1.0
            if np.shape(Qvectors_list[grainindex])[1] != 3:
                raise TypeError
        except IndexError:
            raise IndexError("vec_and_indices has not the proper shape")

        Qsquare = Qx ** 2 + Qy ** 2 + Qz ** 2

        # corresponding Ewald sphere radius for each spots
        # (proportional to photons Energy)
        Rewald = Qsquare / 2.0 / np.abs(Qx)

        # Kf direction selection
        if kf_direction == "Z>0":  # top reflection geometry
            ratiod = detectordistance / Qz
            Ycam = ratiod * (Qx + Rewald)
            Xcam = ratiod * (Qy)
        elif kf_direction == "Y>0":  # side reflection geometry (for detector between the GMT hutch door and the sample (beam coming from right to left)
            ratiod = detectordistance / Qy
            Xcam = ratiod * (Qx + Rewald)
            Ycam = ratiod * (Qz)

        elif kf_direction == "Y<0":  # other side reflection
            ratiod = detectordistance / np.abs(Qy)
            Xcam = -ratiod * (Qx + Rewald)
            Ycam = ratiod * (Qz)
        elif kf_direction == "X>0":  # transmission geometry
            ratiod = detectordistance / np.abs(Qx + Rewald)
            Xcam = -1.0 * ratiod * (Qy)
            Ycam = -ratiod * (Qz)
        elif kf_direction == "X<0":  # back reflection geometry
            ratiod = detectordistance / np.abs(Qx + Rewald)
            Xcam = ratiod * (Qy)
            Ycam = ratiod * (Qz)
        elif kf_direction == "4PI":  # to keep all scattered spots
            Xcam = np.zeros_like(Qy)
            Ycam = np.zeros_like(Qy)

        elif isinstance(kf_direction, (list, np.array)):
            if len(kf_direction) != 2:
                raise ValueError(
                    "kf_direction must be defined by a list of two angles !")
            else:
                Xcam = np.zeros_like(Qy)
                Ycam = np.zeros_like(Qy)
        else:
            raise ValueError("Unknown laue geometry code for kf_direction parameter")

        #print("Xcam, Ycam",Xcam,Ycam)
        # print Ycam
        # print "******************"
        # On camera filter
        # print "detectordiameter", detectordiameter
        halfCamdiametersquare = (detectordiameter / 2.0) ** 2
        # TODO: should contain Xcam-Xcamcen (mm) and Ycam-Ycamcen with Xcamcen, Ycamcen
        # given by user
        onCam_cond = Xcam ** 2 + Ycam ** 2 <= halfCamdiametersquare

        #print('onCam_cond',onCam_cond)
        #print('onCam_cond   true',np.where(onCam_cond==True))
        # resulting arrays
        oncam_Qx = np.compress(onCam_cond, Qx)
        oncam_Qy = np.compress(onCam_cond, Qy)
        oncam_Qz = np.compress(onCam_cond, Qz)

        oncam_R = np.compress(onCam_cond, Rewald)
        oncam_XplusR = oncam_Qx + oncam_R

        # compute 2theta, chi  (in radians)
        oncam_2theta = np.arccos(oncam_XplusR / np.sqrt(oncam_XplusR ** 2 + oncam_Qy ** 2 + oncam_Qz ** 2))

        #print('oncam_2theta',oncam_2theta)
        # be careful of the of sign
        oncam_chi = np.arctan(1.0 * oncam_Qy / oncam_Qz)
        #         oncam_chi = np.arctan2(1. * oncam_Qy, oncam_Qz)
        # TODO: replace by arctan2(1. * oncam_Qy ,oncam_Qz) ??

        #        print "oncam_2theta", oncam_2theta
        #        print "oncam_chi", oncam_chi
        #        print "len(oncam)", len(oncam_2theta)

        # creates spot instances which takes some times...
        # will compute spots on the camera
        # (if HarmonicsRemoval = 0 harmonics still exist)
        # (if HarmonicsRemoval = 1 harmonics are removed, fundamentals are kept)
        # (those of lowest energy)
        if fastcompute == 0:
            try:
                indi_H = HKLs_list[grainindex][:, 0]
                indi_K = HKLs_list[grainindex][:, 1]
                indi_L = HKLs_list[grainindex][:, 2]
                if np.shape(HKLs_list[grainindex])[1] != 3:
                    raise TypeError
            except TypeError:
                emsg = "HKLs_list must be a list of array of 3D vectors"
                raise TypeError(emsg)

            oncam_vec = np.array([oncam_Qx, oncam_Qy, oncam_Qz]).T

            # print('onCam_cond', onCam_cond)
            oncam_H = np.compress(onCam_cond, indi_H)
            oncam_K = np.compress(onCam_cond, indi_K)
            oncam_L = np.compress(onCam_cond, indi_L)
            oncam_HKL = np.transpose(np.array([oncam_H, oncam_K, oncam_L]))

            # build list of spot objects
            listspot = get2ThetaChi_geometry(oncam_vec,
                                            oncam_HKL,
                                            detectordistance=detectordistance,
                                            pixelsize=pixelsize,
                                            dim=dim,
                                            kf_direction=kf_direction)
            #print("listspot", listspot)
            # print("oncam_HKL", oncam_HKL.tolist())
            # Creating list of spot with or without harmonics
            if HarmonicsRemoval and listspot:
                # ListSpots_Oncam_wo_harmonics[grainindex] = RemoveHarmonics(listspot)
                # (oncam_HKL_filtered, toremove)
                (_, toremove) = CP.FilterHarmonics_2(oncam_HKL, return_indices_toremove=1)
                #print('toremove',toremove)
                listspot = np.delete(np.array(listspot), toremove).tolist()

            # feeding final list of spots
            ListSpots_Oncam_wo_harmonics[grainindex] = listspot

            if listspot is not None:
                totalnbspots += len(listspot)

            if fileOK:
                IOLT.Writefile_data_log(ListSpots_Oncam_wo_harmonics[grainindex],
                                        grainindex,
                                        linestowrite=linestowrite)

            # print "Number of spot in camera w / o harmonics",len(ListSpots_Oncam_wo_harmonics[grainindex])

        # (fastcompute = 1) no instantiation of spot object
        # will return 2theta, chi for each grain
        # no harmonics removal
        elif fastcompute == 1:
            Oncam2theta[grainindex] = oncam_2theta
            Oncamchi[grainindex] = oncam_chi

            totalnbspots += len(oncam_2theta)

    if verbose: print('total number of spots for all the %d grain(s) in filterLaueSpots():  '%nbofgrains, totalnbspots)
    if totalnbspots == 0:
        return
    # outputs and returns
    if fileOK:
        linestowrite.append(["\n"])
        linestowrite.append(["--------------------------------------------------------"])
        linestowrite.append(["------------- Simulation Data --------------------------"])
        linestowrite.append(["--------------------------------------------------------"])
        linestowrite.append(["\n"])
        linestowrite.append(["#grain, h, k, l, energy(keV), 2theta (deg), chi (deg), X_Xmas, Y_Xmas, X_JSM, Y_JSM, Xtest, Ytest"])
    if fastcompute == 0:
        # list of elements which are list of objects of spot class (1 element / grain)
        return ListSpots_Oncam_wo_harmonics
    elif fastcompute == 1:
        # list of elements which are composed by two arrays (2theta, chi) (1 element / grain)
        return (np.concatenate(Oncam2theta) / DEG, np.concatenate(Oncamchi) / DEG)


def filterLaueSpots_full_np(veccoord, indicemiller, onlyXYZ=False, HarmonicsRemoval=1,
                                                        fastcompute=0,
                                                        kf_direction=DEFAULT_TOP_GEOMETRY,
                                                        detectordistance=DEFAULT_DETECTOR_DISTANCE,
                                                        detectordiameter=DEFAULT_DETECTOR_DIAMETER,
                                                        pixelsize=165.0 / 2048,
                                                        dim=(2048, 2048),
                                                        grainindex=0):
    r""" Calculates spots data for an individual grain
    on camera and without harmonics
    and on CCD camera

    :param veccoord : list of elements corresponding to 1 grain,
                    each element is composed by: array of q vector [Qx,Qy,Qz]
    :param indicemiller : list of Miller indices [h,k,l] array of indices

    :param HarmonicsRemoval: 1 removes harmonics according to their miller indices
                            (only for fastcompute = 0)

    :param fastcompute:  1 output a list for each grain of 2theta spots and a list for each grain of chi spots
                            (HARMONICS spots are still HERE!)
                           0 output list for each grain of spots with


    :param kf_direction: label for detection geometry
                    (CCD plane with respect to
                    the incoming beam and sample)
    :type kf_direction: string

    :return: tuple of lists of Twtheta Chi Energy Millers if fastcompute=0

                tuple of lists 2theta, chi          if fastcompute=1

    .. note::
        USED in detectorCalibration...simulate_theo
        and tentatively on matching rate)
    """
    VecX = veccoord[:, 0] * 1.0
    VecY = veccoord[:, 1] * 1.0
    VecZ = veccoord[:, 2] * 1.0

    Vecsquare = VecX ** 2 + VecY ** 2 + VecZ ** 2

    # correspondinf Ewald sphere radius for each spots
    # (proportional to photons Energy)
    Rewald = Vecsquare / 2.0 / np.abs(VecX)

    # Kf direction selection
    if kf_direction == "Z>0":  # top reflection geometry
        # VecZ is >0
        ratiod = detectordistance / VecZ
        Ycam = ratiod * (VecX + Rewald)
        Xcam = ratiod * (VecY)
    elif kf_direction == "Y>0":  # side reflection geometry (for detector between the GMT hutch door and the sample (beam coming from right to left)
        ratiod = detectordistance / VecY
        Xcam = ratiod * (VecX + Rewald)
        Ycam = ratiod * (VecZ)
    elif kf_direction == "Y<0":  # other side reflection
        ratiod = detectordistance / np.abs(VecY)
        Xcam = -ratiod * (VecX + Rewald)
        Ycam = ratiod * (VecZ)
    elif kf_direction == "X>0":  # transmission geometry
        ratiod = detectordistance / np.abs(VecX + Rewald)
        Xcam = -1.0 * ratiod * (VecY)
        Ycam = -ratiod * (VecZ)
    elif kf_direction == "X<0":  # back reflection geometry
        # pos0 = np.where(np.abs(VecX + Rewald) < 0.0000000001)
        # print('pos0', pos0)
        # print(VecX[pos0])
        # print(Rewald[pos0])
        # if VecX +Rewald = 0 then kf is // z axis
        # peaks are likely not to be intercepted bt detector plane
        # except if plane is
        ratiod = detectordistance / np.abs(VecX + Rewald)
        Xcam = ratiod * (VecY)
        Ycam = ratiod * (VecZ)
    elif kf_direction == "4PI":  # to keep all scattered spots
        Xcam = np.zeros_like(VecY)
        Ycam = np.zeros_like(VecY)

    elif isinstance(kf_direction, (list, np.array)):
        if len(kf_direction) != 2:
            raise ValueError("kf_direction must be defined by a list of two angles !")
        else:
            Xcam = np.zeros_like(VecY)
            Ycam = np.zeros_like(VecY)
    else:
        raise ValueError("Unknown laue geometry code for kf_direction parameter")

    # print "Xcam, Ycam",Xcam
    # print Ycam
    # ----------------------------------------------------------
    # Approximate selection of spots in camera accroding to Xcam and Ycam
    # onCam_cond  conditions
    # ------------------------------------------------------------

    # print "******************"
    # On camera filter
    # print "detectordiameter", detectordiameter
    halfCamdiametersquare = (detectordiameter / 2.0) ** 2
    # TODO: should contain Xcam-Xcamcen (mm) and Ycam-Ycamcen with Xcamcen, Ycamcen
    # given by user
    onCam_cond = Xcam ** 2 + Ycam ** 2 <= halfCamdiametersquare
    # resulting arrays
    oncam_vecX = np.compress(onCam_cond, VecX)
    oncam_vecY = np.compress(onCam_cond, VecY)
    oncam_vecZ = np.compress(onCam_cond, VecZ)

    if onlyXYZ:
        return np.array([oncam_vecX, oncam_vecY, oncam_vecZ]).T

    # creates spot instances which takes some times...
    # will compute spots on the camera
    # (if HarmonicsRemoval = 0 harmonics still exist)
    # (if HarmonicsRemoval = 1 harmonics are removed, fundamentals are kept)
    # (those of lowest energy)
    if fastcompute == 0:

        indi_H = indicemiller[:, 0]
        indi_K = indicemiller[:, 1]
        indi_L = indicemiller[:, 2]

        oncam_vec = np.array([oncam_vecX, oncam_vecY, oncam_vecZ]).T

        oncam_H = np.compress(onCam_cond, indi_H)
        oncam_K = np.compress(onCam_cond, indi_K)
        oncam_L = np.compress(onCam_cond, indi_L)
        oncam_HKL = np.transpose(np.array([oncam_H, oncam_K, oncam_L]))

        # build list of spot objects
        TwthetaChiEnergyMillers_list_one_grain = get2ThetaChi_geometry_full_np(
                                                                oncam_vec,
                                                                oncam_HKL,
                                                                detectordistance=detectordistance,
                                                                pixelsize=pixelsize,
                                                                dim=dim,
                                                                kf_direction=kf_direction)

        #         print 'TwthetaChiEnergy_list_one_grain', TwthetaChiEnergy_list_one_grain
        # Creating list of spot with or without harmonics
        if HarmonicsRemoval and True:  # listspot:
            raise ValueError("Harmonic removal is not implemented and listspot does not exist anymore")
            #                ListSpots_Oncam_wo_harmonics[i] = RemoveHarmonics(listspot)
            # (oncam_HKL_filtered, toremove) = CP.FilterHarmonics_2(
            #     oncam_HKL, return_indices_toremove=1
            # )
            # listspot = np.delete(np.array(listspot), toremove).tolist()

        # print "Number of spot in camera w / o harmonics",len(ListSpots_Oncam_wo_harmonics[i])

    # (fastcompute = 1) no instantiation of spot object
    # will return 2theta, chi for each grain
    # no harmonics removal
    elif fastcompute == 1:

        oncam_R = np.compress(onCam_cond, Rewald)
        oncam_XplusR = oncam_vecX + oncam_R
        # compute 2theta, chi
        oncam_2theta = (
            np.arccos(oncam_XplusR / np.sqrt(oncam_XplusR ** 2 + oncam_vecY ** 2 + oncam_vecZ ** 2))
            / DEG)
        # be careful of the of sign
        oncam_chi = np.arctan(1.0 * oncam_vecY / oncam_vecZ) / DEG
        #         oncam_chi = np.arctan2(1. * oncam_vecY, oncam_vecZ)
        # TODO: replace by arctan2(1. * oncam_vecY ,oncam_vecZ) ??

    if fastcompute == 0:
        # list spots data
        return TwthetaChiEnergyMillers_list_one_grain
    elif fastcompute == 1:
        # list of elements which are composed by two arrays (2theta, chi) (1 element / grain)
        return oncam_2theta, oncam_chi


def get2ThetaChi_geometry(oncam_vec, oncam_HKL, detectordistance=DEFAULT_DETECTOR_DISTANCE,
                                                                pixelsize=165.0 / 2048,
                                                                dim=(2048, 2048),
                                                                kf_direction=DEFAULT_TOP_GEOMETRY):
    r"""
    computes list of spots instances from oncam_vec (q 3D vectors)
    and oncam_HKL (miller indices 3D vectors)

    :param oncam_vec: q vectors [qx,qy,qz] (corresponding to kf collected on camera)
    :type oncam_vec: array with 3D elements (shape = (n,3))

    :param dim: CCD frame dimensions (nb pixels, nb pixels)
    :type dim: list or tuple of 2 integers

    :param detectordistance: approximate distance detector sample
    :param detectordistance: float or integer

    :param: kf_direction : label for detection geometry
                    (CCD plane with respect to
                    the incoming beam and sample)
    :type: kf_direction: string

    :param pixelsize: pixel size in mm
    :type pixelsize: float

    :return: list of spot instances

    .. note::
        USED in lauecore.filterLaueSpots

    .. todo::
        * to be replaced by something else not using spot class
        * put this function in LaueGeometry module ?
    """
    if len(oncam_vec) != len(oncam_HKL):
        raise ValueError("Wrong input for get2ThetaChi_geometry()")

    listspot = []
    options_createspot = {"allattributes": 0, "pixelsize": pixelsize, "dim": dim}

    dictcase = {"Z>0": create_spot,   # top reflection geom
                "Y>0": create_spot_side_pos,  # side + reflection geom
                "Y<0": create_spot_side_neg,  # side - reflection geom
                "X>0": create_spot_front, # transmission geom
                "X<0": create_spot_back,  # back reflection geom
                "4PI": create_spot_4pi}   # all spots

    for position, indices in zip(oncam_vec, oncam_HKL):
        try:
            function_create_spot = dictcase[kf_direction]
        except TypeError:
            function_create_spot = create_spot_4pi

        spotcreated = function_create_spot(position, indices, detectordistance,
                                                **options_createspot)
        if spotcreated is not None:
            listspot.append(spotcreated)
    return listspot


def get2ThetaChi_geometry_full_np(oncam_vec, oncam_HKL, detectordistance=DEFAULT_DETECTOR_DISTANCE,
                                                                pixelsize=165.0 / 2048,
                                                                dim=(2048, 2048),
                                                                kf_direction=DEFAULT_TOP_GEOMETRY):
    r"""
    computes 2theta chi from oncam_vec (only 3D Q vectors corresponding to reflections on camera)
    and oncam_HKL (miller indices 3D vectors) for all spots of one grain

    :param oncam_vec: q vectors [qx,qy,qz] (corresponding to kf collected on camera)
    :type oncam_vec: array with 3D elements (shape = (n,3))

    :param dim: CCD frame dimensions (nb pixels, nb pixels)
    :type dim: list or tuple of 2 integers

    :param detectordistance: approximate distance detector sample
    :param detectordistance: float or integer

    :param: kf_direction: label for detection geometry
                    (CCD plane with respect to
                    the incoming beam and sample)
    :type: kf_direction: string

    :param pixelsize: pixel size in mm
    :type pixelsize: float

    :return: list of spot

    .. note::
        * USED in lauecore.filterLaueSpots_full_np
        * USED in DetectorCalibration.simultate_theo

    .. todo::
        * Only geometry Z>0 (top reflection) and X>0 (transmission) are vectorized by numpy
        * TODO: put this function obviously in find2thetachi ?
    """
    if len(oncam_vec) != len(oncam_HKL):
        raise ValueError("Wrong input for get2ThetaChi_geometry_full_np()")

    # TwthetaChiEnergy_list = []
    options_createspot = {"allattributes": 0, "pixelsize": pixelsize, "dim": dim}

    dictcase = {"Z>0": create_spot_np,
                "Y>0": create_spot_side_pos,
                "Y<0": create_spot_side_neg,
                "X>0": create_spot_np,
                "X<0": create_spot_np, #create_spot_back,
                "4PI": create_spot_4pi}

    try:
        function_create_spot = dictcase[kf_direction]
    except TypeError:
        function_create_spot = create_spot_4pi

    TwthetaChiEnergyMillers_list = function_create_spot(
        oncam_vec, oncam_HKL, detectordistance, **options_createspot)
    return TwthetaChiEnergyMillers_list


def RemoveHarmonics(listspot):
    r"""
    removes harmonics present in listspot (list of objects of spot class)

    .. todo:: NOT USED ANYMORE!
    """
    _invdict = {}

    # print "len after sorting wtih respect to __hash__()",len(sorted(listspot, reverse = True))
    for elem in sorted(listspot, reverse=True):
        # print "elem, elem.__hash__()",elem.Millers, elem.__hash__()

        _invdict[elem.__hash__()] = elem
        _oncamsansh = [_invdict[cle] for cle in list(_invdict.keys())]

    # print "Number of fundamental spots (RS directions): %d"%len(_oncamsansh)
    return _oncamsansh


def calcSpots_fromHKLlist(UB, B0, HKL, dictCCD):
    r"""
    computes all Laue Spots properties on 2D detector from a list of hkl
    (given structure by B0 matrix, orientation by UB matrix, and detector geometry by dictCCD)

    :param UB: orientation matrix (rotation -and if any- strain)
    :type UB: 3x3 array (or list)

    :param B0: initial a*,b*,c* reciprocal unit cell basis vector in Lauetools frame (x// ki))
    :type B0: 3x3 array (or list)

    :param HKL: array of Miller indices
    :type HKL: array with shape = (n,3)

    :param dictCCD: dictionnary of CCD properties (with key 'CCDparam', 'pixelsize','dim')
        for 'ccdparam' 5 CCD calibration parameters [dd,xcen,ycen,xbet,xgam], pixelsize in mm, and (dim1, dim2)
    :param dictCCD: dict object

    :returns: list of arrays H, K, L, Qx, Qy, Qz, X, Y, twthe, chi, Energy

    Fundamental equation
    :math:`{\bf q} = UB*B0 * {\bf G^*}`
    with :math:`{\bf G^*} = h{\bf a^*}+k{\bf b^*}+l{\bf c^*}`

    .. note::
        USED in DetectorCalibration.OnWriteResults, and PlotRefineGUI.onWriteFitFile
    """

    detectorparam = dictCCD["CCDparam"]
    pixelsize = dictCCD["pixelsize"]
    # dim = dictCCD["dim"]
    if "kf_direction" in dictCCD:
        kf_direction = dictCCD["kf_direction"]
    else:
        kf_direction = "Z>0"

    # H,K,L
    tHKL = np.transpose(HKL)

    # initial lattice rotation and distorsion

    tQ = np.dot(np.dot(UB, B0), tHKL)
    # results are qx,qy,qz

    # Q**2
    Qsquare = np.sum(tQ ** 2, axis=0)
    # norms of Q vectors
    Qn = 1.0 * np.sqrt(Qsquare)

    twthe, chi = LTGeo.from_qunit_to_twchi(tQ / Qn)

    X, Y, _ = LTGeo.calc_xycam_from2thetachi(twthe,
                                                chi,
                                                detectorparam,
                                                verbose=0,
                                                pixelsize=pixelsize,
                                                kf_direction=kf_direction)

    # E = (C)*(-q**2/qx/2)
    Energy = (CST_ENERGYKEV) * (-0.5 * Qsquare / tQ[0])

    # theoretical values
    H, K, L = tHKL
    Qx, Qy, Qz = tQ

    return H, K, L, Qx, Qy, Qz, X, Y, twthe, chi, Energy


def emptylists(n):
    r"""
    builds a list of n empty lists : [[],[],[], ...,[]]
    """
    return [[] for k in list(range(n))]


def SimulateLaue_merge( grains, emin, emax, detectorparameters, only_2thetachi=True,
                                                output_nb_spots=False,
                                                kf_direction=DEFAULT_TOP_GEOMETRY,
                                                ResolutionAngstrom=False,
                                                removeharmonics=0,
                                                pixelsize=165 / 2048.0,
                                                dim=(2048, 2048),
                                                detectordiameter=None,
                                                dictmaterials=dict_Materials):
    r"""
    Simulates Laue pattern full data from a list of grains and concatenate results data

    :param grains: list of 4 elements grain parameters

    :param only_2thetachi: * True, return only concatenated grains data 2theta and chi,
                           * False, return All_Twicetheta, All_Chi, All_Miller_ind,
                                    All_posx, All_posy, All_Energy

    :param output_nb_spots: * True, output a second element (in addition to data)
                            with list of partial nb of spots per grain
                            (to know the grain origin of spots)
    """
    # use SimulateLaue
    if not only_2thetachi:
        All_Twicetheta = []
        All_Chi = []
        All_Miller_ind = []
        All_posx = []
        All_posy = []
        All_Energy = []
        All_nb_spots = []

        for grain in grains:
            (Twicetheta, Chi, Miller_ind, posx, posy, Energy) = SimulateLaue(
                grain,
                emin,
                emax,
                detectorparameters,
                kf_direction=kf_direction,
                ResolutionAngstrom=ResolutionAngstrom,
                removeharmonics=removeharmonics,
                pixelsize=pixelsize,
                dim=dim,
                detectordiameter=detectordiameter,
                dictmaterials=dictmaterials)

            nb_spots = len(Twicetheta)

            All_Twicetheta.append(Twicetheta)
            All_Chi.append(Chi)
            All_Miller_ind.append(Miller_ind)
            All_posx.append(posx)
            All_posy.append(posy)
            All_Energy.append(Energy)
            All_nb_spots.append(nb_spots)

        All_Twicetheta = np.concatenate(All_Twicetheta)
        All_Chi = np.concatenate(All_Chi)
        All_Miller_ind = np.concatenate(All_Miller_ind)
        All_posx = np.concatenate(All_posx)
        All_posy = np.concatenate(All_posy)
        All_Energy = np.concatenate(All_Energy)

        toreturn = (All_Twicetheta,
                    All_Chi,
                    All_Miller_ind,
                    All_posx,
                    All_posy,
                    All_Energy)

    # Use SimulateResult
    else:
        simulparameters = {}
        simulparameters["detectordiameter"] = detectordiameter
        simulparameters["kf_direction"] = kf_direction
        simulparameters["detectordistance"] = detectorparameters[0]
        simulparameters["pixelsize"] = pixelsize

        All_TwicethetaChi = []
        All_nb_spots = []
        for grain in grains:
            TwicethetaChi = SimulateResult(grain,
                                            emin,
                                            emax,
                                            simulparameters,
                                            fastcompute=1,
                                            ResolutionAngstrom=False,
                                            dictmaterials=dictmaterials)
            nb_spots = len(TwicethetaChi[0])

            All_TwicethetaChi.append(TwicethetaChi)
            All_nb_spots.append(nb_spots)

        All_TwicethetaChi = np.concatenate(All_TwicethetaChi)

        toreturn = All_TwicethetaChi

    if output_nb_spots:
        return toreturn, All_nb_spots
    else:
        return toreturn


def SimulateLaue_twins(grainparent, twins_operators, emin, emax, detectorparameters,
                                                                only_2thetachi=True,
                                                                output_nb_spots=False,
                                                                kf_direction=DEFAULT_TOP_GEOMETRY,
                                                                ResolutionAngstrom=False,
                                                                removeharmonics=0,
                                                                pixelsize=165 / 2048.0,
                                                                dim=(2048, 2048),
                                                                detectordiameter=None,
                                                                dictmaterials=dict_Materials):
    r"""
    Simulates Laue pattern full data for twinned grain

    :param grainparent: list of 4 elements grain parameter

    :param twins_operators: list of 3*3 matrices corresponding of Matrices

    :param output_nb_spots: True, output a second element with list of partial nb of spots per grain

    .. note:: USED in test only in detectorCalibration...simulate_theo  to simulate 2 twinned crystals
    """
    # nb_twins = len(twins_operators)

    Bmat, dilat, Umat, extinction = grainparent

    grains = [grainparent]

    for _, twin_op in enumerate(twins_operators):
        twinUmat = np.dot(Umat, twin_op)
        grains.append([Bmat, dilat, twinUmat, extinction])

    return SimulateLaue_merge(grains, emin, emax, detectorparameters, only_2thetachi=only_2thetachi,
                                                            output_nb_spots=output_nb_spots,
                                                            kf_direction=kf_direction,
                                                            ResolutionAngstrom=ResolutionAngstrom,
                                                            removeharmonics=removeharmonics,
                                                            pixelsize=pixelsize,
                                                            dim=dim,
                                                            detectordiameter=detectordiameter,
                                                            dictmaterials=dictmaterials)


def SimulateLaue(grain, emin, emax, detectorparameters, kf_direction=DEFAULT_TOP_GEOMETRY,
                                                            ResolutionAngstrom=False,
                                                            removeharmonics=0,
                                                            pixelsize=165 / 2048.0,
                                                            dim=(2048, 2048),
                                                            detectordiameter=None,
                                                            force_extinction=None,
                                                            dictmaterials=dict_Materials,
                                                            version=1):
    r"""Computes Laue Pattern spots positions, scattering angles, miller indices
                            for a SINGLE grain or Xtal

    :param grain: crystal parameters made of a 4 elements list
    :param emin: minimum bandpass energy (keV)
    :param emax: maximum bandpass energy (keV)

    :param removeharmonics:
        * 1, removes harmonics spots and keep fondamental spots (or reciprocal direction)
            (with lowest Miller indices)

        * 0 keep all spots (including harmonics)

    :return: single grain data: Twicetheta, Chi, Miller_ind, posx, posy, Energy

    .. todo::
        To update to accept kf_direction not only in reflection geometry

    .. note::
        USED in detectorCalibration...simulate_theo  for non routine geometry (ie except Z>0 (reflection top) X>0 (transmission)
    """

    if detectordiameter is None:
        DETECTORDIAMETER = pixelsize * dim[0]
    else:
        DETECTORDIAMETER = detectordiameter

    key_material = grain[3]

    grain = CP.Prepare_Grain(
        key_material, grain[2], force_extinction=force_extinction, dictmaterials=dictmaterials)

    Spots2pi = getLaueSpots(CST_ENERGYKEV / emax,
                            CST_ENERGYKEV / emin,
                            [grain],
                            fastcompute=0,
                            verbose=0,
                            kf_direction=kf_direction,
                            ResolutionAngstrom=ResolutionAngstrom,
                            dictmaterials=dictmaterials)

    #     print "len Spots2pi", len(Spots2pi[0][0])

    # list of spot which are on camera (with harmonics)
    ListofListofSpots = filterLaueSpots(Spots2pi,
                                            fileOK=0,
                                            fastcompute=0,
                                            detectordistance=detectorparameters[0],
                                            detectordiameter=DETECTORDIAMETER,
                                            kf_direction=kf_direction,
                                            HarmonicsRemoval=removeharmonics,
                                            pixelsize=pixelsize)

    ListofSpots = ListofListofSpots[0]

    Twicetheta = np.array([spot.Twicetheta for spot in ListofSpots])
    Chi = np.array([spot.Chi for spot in ListofSpots])
    Miller_ind = np.array([list(spot.Millers) for spot in ListofSpots])
    Energy = np.array([spot.EwaldRadius * CST_ENERGYKEV for spot in ListofSpots])

    posx, posy = LTGeo.calc_xycam_from2thetachi(Twicetheta,
                                                Chi,
                                                detectorparameters,
                                                verbose=0,
                                                pixelsize=pixelsize,
                                                kf_direction=kf_direction, version=version)[:2]

    return Twicetheta, Chi, Miller_ind, posx, posy, Energy


def SimulateLaue_full_np(grain, emin, emax, detectorparameters, kf_direction=DEFAULT_TOP_GEOMETRY,
                                                                ResolutionAngstrom=False,
                                                                removeharmonics=0,
                                                                pixelsize=165 / 2048.0,
                                                                dim=(2048, 2048),
                                                                detectordiameter=None,
                                                                force_extinction=None,
                                                                dictmaterials=dict_Materials):
    r"""Compute Laue Pattern spots positions, scattering angles, miller indices
                            for a SINGLE grain or Xtal using numpy vectorization

    :param grain:    crystal parameters in a 4 elements list
    :param emin: minimum bandpass energy (keV)
    :param emax: maximum bandpass energy (keV)

    :param removeharmonics: 1, remove harmonics spots and keep fondamental spots
                            (with lowest Miller indices)

    :return: single grain data: Twicetheta, Chi, Miller_ind, posx, posy, Energy

    .. todo::
        update to accept kf_direction not only in reflection geometry

    .. note::
        USED in detectorCalibration...simulate_theo for routine geometry Z>0 (reflection top) X>0 (transmission)
    """

    if detectordiameter is None:
        DETECTORDIAMETER = pixelsize * dim[0]
    else:
        DETECTORDIAMETER = detectordiameter
    # use DEFAULT_TOP_GEOMETRY <=> kf_direction = 'Z>0'

    key_material = grain[3]
    grain = CP.Prepare_Grain(key_material, grain[2],
                                    force_extinction=force_extinction, dictmaterials=dictmaterials)

    Qxyz, HKL = getLaueSpots(CST_ENERGYKEV / emax,
                            CST_ENERGYKEV / emin,
                            [grain],
                            fastcompute=0,
                            verbose=0,
                            kf_direction=kf_direction,
                            ResolutionAngstrom=ResolutionAngstrom,
                            dictmaterials=dictmaterials)

    # list of spot which are on camera (with harmonics)
    TwthetaChiEnergyMillers_list_one_grain_wo_harmonics = filterLaueSpots_full_np(
                                                Qxyz[0],
                                                HKL[0],
                                                fastcompute=0,
                                                detectordistance=detectorparameters[0],
                                                detectordiameter=DETECTORDIAMETER,
                                                kf_direction=kf_direction,
                                                HarmonicsRemoval=0,
                                                pixelsize=pixelsize)

    Twicetheta, Chi, Energy, Miller_ind = TwthetaChiEnergyMillers_list_one_grain_wo_harmonics[:4]

    posx, posy = LTGeo.calc_xycam_from2thetachi(Twicetheta,
                                                Chi,
                                                detectorparameters,
                                                verbose=0,
                                                pixelsize=pixelsize,
                                                kf_direction=kf_direction)[:2]

    if removeharmonics:
        # remove harmonics:
        _, _, tokeep = GT.removeClosePoints(posx, posy, 0.05)
        print('removeharmonics = 1 in SimulateLaue_full_np() tokeep',tokeep)

        s_tth = Twicetheta[tokeep]
        s_chi = Chi[tokeep]
        s_miller_ind = Miller_ind[tokeep]
        s_posx = posx[tokeep]
        s_posy = posy[tokeep]
        s_E = Energy[tokeep]
        return s_tth, s_chi, s_miller_ind, s_posx, s_posy, s_E
    else:
        return Twicetheta, Chi, Miller_ind, posx, posy, Energy


def SimulateResult(grain, emin, emax, simulparameters,
                    fastcompute=1, ResolutionAngstrom=False, dictmaterials=dict_Materials):
    r"""Simulates 2theta chi of Laue Pattern spots for ONE SINGLE grain

    :param grain: crystal parameters in a 4 elements list
    :param emin: minimum bandpass energy (keV)
    :param emax: maximum bandpass energy (keV)

    :return: 2theta, chi

    .. warning:: Need of approximate detector distance and diameter to restrict simulation to a limited solid angle

    .. note::
        * USED: in AutoindexationGUI.OnStart, LaueToolsGUI.OnCheckOrientationMatrix
        * USED also IndexingImageMatching, lauecore.SimulateLaue_merge
    """

    detectordiameter = simulparameters["detectordiameter"]
    kf_direction = simulparameters["kf_direction"]
    detectordistance = simulparameters["detectordistance"]
    pixelsize = simulparameters["pixelsize"]

    # PATCH: redefinition of grain to simulate any unit cell(not only cubic)
    key_material = grain[3]
    grain = CP.Prepare_Grain(key_material, grain[2], dictmaterials=dictmaterials)
    # -----------------------------------------------------------------------------

    # print "grain in SimulateResult()",grain

    spots2pi = getLaueSpots(CST_ENERGYKEV / emax,
                            CST_ENERGYKEV / emin,
                            [grain],
                            fastcompute=fastcompute,
                            ResolutionAngstrom=ResolutionAngstrom,
                            verbose=0,
                            kf_direction=kf_direction,
                            dictmaterials=dictmaterials)
    # ---------------------------------------------------------------------------

    # array(vec) and array(indices)  of spots exiting the crystal in 2pi steradian
    # if fastcompute = 0 array(indices) = 0 and TwicethetaChi is a list of spot object
    # k_direction =(Z>0)
    # TODO: to be argument if camera is far from preset kf_direction!!
    # spots2pi = LAUE.generalfabriquespot_fromMat_veryQuick(CST_ENERGYKEV/emax, CST_ENERGYKEV/emin,[grain],1,
    # fastcompute = fastcompute, fileOK = 0, verbose = 0, kf_direction = 'Z>0')

    # 2theta, chi of spot which are on camera(without harmonics)
    TwicethetaChi = filterLaueSpots(spots2pi,
                                    fileOK=0,
                                    fastcompute=fastcompute,
                                    detectordistance=detectordistance,
                                    detectordiameter=detectordiameter * 1.2,
                                    kf_direction=kf_direction,
                                    pixelsize=pixelsize)

    return TwicethetaChi

def B_DebyeWaller(U):
    """ compute B term of exp Debye waller factor
    exp - B  (sintheta/lambda)
    with U mean square displacement
    """
    return 8 * np.pi**2 * U


def StructureFactorCubic(h, k, l, extinctions="dia"):
    """
    computes structure factor of cubic
    """
    assert extinctions == 'dia'
    pi = np.pi
    F = (1 + np.exp(-1.0j * pi / 2.0 * (h + k + l))) * (1 + np.exp(-1.0j * pi * (k + l))
                                                        + np.exp(-1.0j * pi * (h + l))
                                                        + np.exp(-1.0j * pi * (h + k)))
    return F


def StructureFactorUO2(h, k, l, qvector, U_U, U_O):
    """
    computes structure factor of CaF2 flurine type structure  SG 225 Fm-3m
    """
    # CaF2 structure
    pi = np.pi
    # q = 4pi*sintheta/lambda
    sol = qvector/(4*pi)
    B_U = B_DebyeWaller(U_U)
    B_O = B_DebyeWaller(U_O)
    fu = atomicformfactor(qvector, "U")*np.exp(-B_U * sol**2)
    fo = atomicformfactor(qvector, "O")*np.exp(-B_O * sol**2)
    F = 4*(fu + 2 * fo * np.cos(pi / 2*((h + k + l)))) * (1 + np.exp(1.0j * pi * (k + l))
                                                        + np.exp(1.0j * pi * (l + h))
                                                        + np.exp(1.0j * pi * (h + k)))
    return F


def atomicformfactor(q, element="Ge"):
    """
    Computes x-ray atomic scattering factor following
    http://lampx.tugraz.at/~hadley/ss1/crystaldiffraction/atomicformfactors/formfactors.php

    :param q: q vector norm in Angst-1
    :returns: scalar, f(q)

    """
    if element == "Ge":
        p = (16.0816, 2.8509, 6.3747, 0.2516, 3.7068, 11.4468, 3.683, 54.7625, 2.1313)

    elif element == "U":
        p = (5.3715, 0.516598, 22.5326, 3.05053, 12.0291, 12.5723, 4.79840, 23.4582, 13.2671)

    elif element == "O":
        p = (3.04850, 13.2771, 2.28680, 5.70110, 1.54630, 0.323900, 0.867000, 32.9089, 0.250800)

    val = 0
    for k in list(range(4)):
        val += p[2 * k] * np.exp(-p[2 * k + 1] * (q / 4 / np.pi) ** 2)
    val += p[-1]
    return val


def simulatepurepattern_np(grain, emin, emax, kf_direction, data_filename, PlotLaueDiagram=1,
                                                        Plot_Data=0,
                                                        verbose=0,
                                                        detectordistance=DEFAULT_DETECTOR_DISTANCE,
                                                        ResolutionAngstrom=False,
                                                        Display_label=1,
                                                        HarmonicsRemoval=1,
                                                        dictmaterials=dict_Materials):
    """
    .. warning:: In test. NOT USED anywhere !!???
    """

    vecind = getLaueSpots(CST_ENERGYKEV / emax, CST_ENERGYKEV / emin, grain,
                    fastcompute=0, kf_direction=kf_direction,
                    verbose=verbose, ResolutionAngstrom=ResolutionAngstrom,
                    dictmaterials=dictmaterials)

    print("len(vecind[0])", len(vecind[0][0]))

    # selecting RR nodes without harmonics (fastcompute = 1 loses the miller indices and RR positions associations for quicker computation)

    oncam_sansh = filterLaueSpots_full_np(vecind[0], vecind[1],
                                        fastcompute=0,
                                        kf_direction=kf_direction,
                                        detectordistance=detectordistance,
                                        HarmonicsRemoval=HarmonicsRemoval)

    return True
